# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import json
from collections import defaultdict
from collections.abc import Iterator
from pathlib import Path
from urllib.parse import urlparse

import shortuuid

from pixano.core import BBox, CompressedRLE, Image
from pixano.data.importers.importer import Importer
from pixano.data.item.item_feature import FeaturesValues, FeatureValues
from pixano.utils import image_to_thumbnail, natural_key


class COCOImporter(Importer):
    """Importer class for COCO instances format datasets

    Attributes:
        info (DatasetInfo): Dataset information
        input_dirs (dict[str, Path]): Dataset input directories
    """

    def __init__(
        self,
        name: str,
        description: str,
        input_dirs: dict[str, Path],
        splits: list[str],
        media_fields: dict[str, str] = None,
    ):
        """Initialize COCO Importer

        Args:
            name (str): Dataset name
            description (str): Dataset description
            input_dirs (dict[str, Path]): Dataset input directories
            splits (list[str]): Dataset splits
            media_fields (dict[str, str]): Dataset media fields, with field names as keys and field types as values. Default to None.
        """

        # Object fields
        object_fields = {
            "original_id": "str",
            "bbox": "bbox",
            "mask": "compressedrle",
            "category": "str",
            "category_id": "int",
            "supercategory": "str",
        }

        # Create features_values
        features_values = FeaturesValues(
            objects={
                "category": FeatureValues(restricted=False, values=[]),
                "category_id": FeatureValues(restricted=False, values=[]),
                "supercategory": FeatureValues(restricted=False, values=[]),
            }
        )
        for split in splits:
            with open(
                input_dirs["objects"] / f"instances_{split}.json", "r", encoding="utf-8"
            ) as f:
                coco_instances = json.load(f)
                if "categories" in coco_instances:
                    for category in coco_instances["categories"]:
                        features_values.objects["category"].values.append(
                            category["name"]
                        )
                        features_values.objects["category_id"].values.append(
                            category["id"]
                        )
                        if "supercategory" in category:
                            features_values.objects["supercategory"].values.append(
                                category["supercategory"]
                            )
        if len(features_values.objects["supercategory"].values) == 0:
            features_values.objects.pop("supercategory")
            object_fields.pop("supercategory")
        else:
            features_values.objects["supercategory"].values = list(
                set(features_values.objects["supercategory"].values)
            )
        FeaturesValues.model_validate(features_values)

        # Create tables
        tables = super().create_tables(
            media_fields,
            object_fields,
        )

        # Initialize Importer
        self.input_dirs = input_dirs
        super().__init__(name, description, tables, splits, features_values)

    def import_rows(self) -> Iterator:
        """Process dataset rows for import

        Yields:
            Iterator: Processed rows
        """

        # Iterate on splits
        for split in self.info.splits:
            # Open annotation files
            with open(
                self.input_dirs["objects"] / f"instances_{split}.json",
                "r",
                encoding="utf-8",
            ) as f:
                coco_instances = json.load(f)

            # Group annotations by image ID
            annotations = defaultdict(list)
            for ann in coco_instances["annotations"]:
                annotations[ann["image_id"]].append(ann)

            # Create a COCO category id to COCO category name and supercategory dictionary
            categories = {}
            supercategories = {}
            for cat in coco_instances["categories"]:
                categories[cat["id"]] = cat["name"]
                supercategories[cat["id"]] = (
                    cat["supercategory"] if "supercategory" in cat else None
                )

            # Process rows
            for im in sorted(
                coco_instances["images"], key=lambda x: natural_key(str(x["id"]))
            ):
                # Load image annotations
                im_anns = annotations[im["id"]]
                # Load image
                file_name_uri = urlparse(im["file_name"])
                if file_name_uri.scheme == "":
                    im_path = self.input_dirs["image"] / split / im["file_name"]
                else:
                    im_path = Path(file_name_uri.path)

                # Create image thumbnail
                im_thumb = image_to_thumbnail(im_path.read_bytes())

                # Set image URI
                im_uri = f"image/{split}/{im_path.name}"

                # Set unique id
                item_id = shortuuid.uuid()

                # Return rows
                rows = {
                    "main": {
                        "db": [
                            {
                                "id": item_id,
                                "original_id": str(im["id"]),
                                "views": ["image"],
                                "split": split,
                            }
                        ]
                    },
                    "media": {
                        "image": [
                            {
                                "id": item_id,
                                "image": Image(im_uri, None, im_thumb).to_dict(),
                            }
                        ]
                    },
                    "objects": {
                        "objects": [
                            {
                                "id": shortuuid.uuid(),
                                "original_id": str(ann["id"]),
                                "item_id": item_id,
                                "view_id": "image",
                                "bbox": (
                                    BBox.from_xywh(ann["bbox"])
                                    .normalize(im["height"], im["width"])
                                    .to_dict()
                                    if ann["bbox"]
                                    else None
                                ),
                                "mask": (
                                    CompressedRLE.encode(
                                        ann["segmentation"], im["height"], im["width"]
                                    ).to_dict()
                                    if ann["segmentation"]
                                    else None
                                ),
                                "category": str(categories[ann["category_id"]]),
                                "category_id": ann["category_id"],
                                "supercategory": supercategories[ann["category_id"]],
                            }
                            for ann in im_anns
                        ]
                    },
                }

                # Remove supercategory if empty
                for obj in rows["objects"]["objects"]:
                    if obj["supercategory"] is None:
                        obj.pop("supercategory")

                yield rows



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path
from typing import Optional

import pyarrow as pa
from pydantic import BaseModel
from s3path import S3Path

from pixano.core import Image, is_image_type
from pixano.data.item.item_feature import ItemFeature


class ItemView(BaseModel):
    """View type for DatasetItem

    Attributes:
        id (str): View ID
        type (str): View type ("image", "video", "point_cloud")
        url (str): View URI
        thumbnail (str, optional): View thumbnail as base 64 URL
        frame_number (int, optional): View frame number
        total_frames (int, optional): View total frames
        features (dict[str, ItemFeature], optional): View features
    """

    id: str
    type: str
    uri: str
    thumbnail: Optional[str] = None
    frame_number: Optional[int] = None
    total_frames: Optional[int] = None
    features: Optional[dict[str, ItemFeature]] = None

    @staticmethod
    def from_pyarrow(
        table: pa.Table,
        schema: pa.schema,
        media_dir: Path,
        media_features: bool = False,
    ) -> dict[str, "ItemView"]:
        """Create dictionary of ItemView from PyArrow Table

        Args:
            table (dict[str, Any]): PyArrow table
            schema (pa.schema): PyArrow schema
            media_dir (Path): Dataset media directory
            media_features (bool, optional): Load media features like image width and height (slow for large item batches)

        Returns:
            dict[ItemView]: Dictionary of ItemView
        """

        # NOTE: Potential change to flattened view fields with one row per view
        item = table.to_pylist()[0] if len(table.to_pylist()) > 0 else None
        views = {}

        # Iterate on fields
        for field in schema:
            # Image
            if is_image_type(field.type):
                if item is not None:
                    im = (
                        item[field.name]
                        if isinstance(item[field.name], Image)
                        else Image.from_dict(item[field.name])
                    )
                    im.uri_prefix = media_dir.absolute().as_uri()
                    api_uri = (
                        (media_dir / im.uri).get_presigned_url()
                        if isinstance(media_dir, S3Path)
                        else f"data/{media_dir.parent.name}/media/{im.uri}"
                    )
                    image_view = ItemView(
                        id=field.name,
                        type="image",
                        uri=api_uri,
                        thumbnail=im.preview_url,
                    )
                    image_view.features = {}
                    if media_features:
                        image_view.features["width"] = ItemFeature(
                            name="width",
                            dtype="int",
                            value=im.width,
                        )

                        image_view.features["height"] = ItemFeature(
                            name="height",
                            dtype="int",
                            value=im.height,
                        )
                    views[field.name] = image_view
                else:
                    views[field.name] = ItemView(id=field.name, type="image", uri="")

            # NOTE: Future support for videos and 3D point clouds

        return views



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException, Response
from fastapi_pagination import Page, Params
from fastapi_pagination.api import create_page, resolve_params

from pixano.data import Dataset, DatasetItem, Settings, get_settings

router = APIRouter(tags=["items"], prefix="/datasets/{ds_id}")


@router.get("/items", response_model=Page[DatasetItem])
async def get_dataset_items(
    ds_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
    params: Params = Depends(),
) -> Page[DatasetItem]:
    """Load dataset items

    Args:
        ds_id (str): Dataset ID
        params (Params, optional): Pagination parameters (offset and limit). Defaults to Depends().

    Returns:
        Page[DatasetItem]: Dataset items page
    """

    # Load dataset
    dataset = Dataset.find(ds_id, settings.data_dir)

    if dataset:
        # Get page parameters
        params = resolve_params(params)
        raw_params = params.to_raw_params()
        total = dataset.num_rows

        # Check page parameters
        start = raw_params.offset
        stop = min(raw_params.offset + raw_params.limit, total)
        if start >= stop:
            raise HTTPException(
                status_code=404,
                detail=f"Invalid page parameters (start {start}, stop {stop})",
            )

        # Load dataset items
        items = dataset.load_items(raw_params.limit, raw_params.offset)

        # Return dataset items
        if items:
            return create_page(items, total=total, params=params)
        raise HTTPException(
            status_code=404,
            detail=f"No items found with page parameters (start {start}, stop {stop}) in dataset",
        )
    raise HTTPException(
        status_code=404,
        detail=f"Dataset {ds_id} not found in {settings.data_dir.absolute()}",
    )


@router.post("/search", response_model=Page[DatasetItem])
async def search_dataset_items(
    ds_id: str,
    query: dict[str, str],
    settings: Annotated[Settings, Depends(get_settings)],
    params: Params = Depends(),
) -> Page[DatasetItem]:
    """Load dataset items with a query

    Args:
        ds_id (str): Dataset ID
        query (dict[str, str]): Search query
        params (Params, optional): Pagination parameters (offset and limit). Defaults to Depends().

    Returns:
        Page[DatasetItem]: Dataset items page
    """

    # Load dataset
    dataset = Dataset.find(ds_id, settings.data_dir)

    if dataset:
        # Get page parameters
        params = resolve_params(params)
        raw_params = params.to_raw_params()
        total = dataset.num_rows

        # Check page parameters
        start = raw_params.offset
        stop = min(raw_params.offset + raw_params.limit, total)
        if start >= stop:
            raise HTTPException(status_code=404, detail="Invalid page parameters")

        # Load dataset items
        items = dataset.search_items(raw_params.limit, raw_params.offset, query)

        # Return dataset items
        if items:
            return create_page(items, total=total, params=params)
        raise HTTPException(
            status_code=404, detail=f"No items found for query '{query}' in dataset"
        )
    raise HTTPException(
        status_code=404,
        detail=f"Dataset {ds_id} not found in {settings.data_dir.absolute()}",
    )


@router.get("/items/{item_id}", response_model=DatasetItem)
async def get_dataset_item(
    ds_id: str,
    item_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetItem:
    """Load dataset item

    Args:
        ds_id (str): Dataset ID
        item_id (str): Item ID

    Returns:
        DatasetItem: Dataset item
    """

    # Load dataset
    dataset = Dataset.find(ds_id, settings.data_dir)

    if dataset:
        # Load dataset item
        item = dataset.load_item(item_id, load_objects=True)

        # Return dataset item
        if item:
            return item
        raise HTTPException(
            status_code=404,
            detail=f"Item '{item_id}' not found in dataset",
        )
    raise HTTPException(
        status_code=404,
        detail=f"Dataset {ds_id} not found in {settings.data_dir.absolute()}",
    )


@router.post("/items/{item_id}", response_model=DatasetItem)
async def post_dataset_item(
    ds_id: str,
    item: DatasetItem,
    settings: Annotated[Settings, Depends(get_settings)],
):
    """Save dataset item

    Args:
        ds_id (str): Dataset ID
        item (DatasetItem): Item to save
    """

    # Load dataset
    dataset = Dataset.find(ds_id, settings.data_dir)

    if dataset:
        # Save dataset item
        dataset.save_item(item)

        # Return response
        return Response()
    raise HTTPException(
        status_code=404,
        detail=f"Dataset {ds_id} not found in {settings.data_dir.absolute()}",
    )


@router.get(
    "/items/{item_id}/embeddings/{model_id}",
    response_model=DatasetItem,
)
async def get_item_embeddings(
    ds_id: str,
    item_id: str,
    model_id: str,
    settings: Annotated[Settings, Depends(get_settings)],
) -> DatasetItem:
    """Load dataset item embeddings

    Args:
        ds_id (str): Dataset ID
        item_id (str): Item ID
        model_id (str): Model ID (ONNX file path)
    """

    # Load dataset
    dataset = Dataset.find(ds_id, settings.data_dir)

    if dataset:
        item = dataset.load_item(
            item_id,
            load_media=False,
            load_active_learning=False,
            load_embeddings=True,
            model_id=model_id,
        )

        # Return dataset item embeddings
        if item:
            return item
        raise HTTPException(
            status_code=404,
            detail=f"No embeddings found for item '{item_id}' with model '{model_id}' in dataset",
        )
    raise HTTPException(
        status_code=404,
        detail=f"Dataset {ds_id} not found in {settings.data_dir.absolute()}",
    )



# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi_pagination.api import add_pagination
from s3path import S3Path

from pixano.app.api import datasets, items, models
from pixano.data.settings import Settings


def create_app(settings: Settings = Settings()) -> FastAPI:
    """Run Pixano app

    Args:
        settings (Settings): App settings

    Returns:
        FastAPI: Pixano app
    """

    # Create app
    app = FastAPI()
    app.add_middleware(
        CORSMiddleware,
        allow_credentials=True,
        allow_methods=["*"],
        allow_headers=["*"],
    )

    # Mount folders
    if isinstance(settings.data_dir, S3Path):
        # If S3, mount models parent folder
        # Check if folder exists
        if not settings.model_dir.exists():
            raise FileNotFoundError(
                f"Local model directory '{settings.model_dir.absolute()}' not found"
            )
        # Mount
        app.mount(
            "/data",
            StaticFiles(directory=settings.model_dir.parent),
            name="data",
        )
    else:
        # If local, mount datasets folder with models subfolder
        # Check if folder exists
        if not settings.data_dir.exists():
            raise FileNotFoundError(
                f"Dataset library '{settings.data_dir.absolute()}' not found"
            )
        # Create models subfolder in case it doesn't exist yet
        settings.model_dir.mkdir(exist_ok=True)
        # Mount
        app.mount(
            "/data",
            StaticFiles(directory=settings.data_dir),
            name="data",
        )

    app.include_router(datasets.router)
    app.include_router(items.router)
    app.include_router(models.router)

    add_pagination(app)
    return app



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import numpy as np
import pyarrow as pa
from pydantic import BaseModel

from pixano.core import (
    BBox,
    BBoxType,
    Camera,
    CameraType,
    CompressedRLE,
    CompressedRLEType,
    DepthImage,
    DepthImageType,
    GtInfo,
    GtInfoType,
    Image,
    ImageType,
    Pose,
    PoseType,
)


def field_to_python(field: str) -> type:
    """Return Python type from string field

    Args:
        field (str): String field

    Returns:
        type: Python type
    """

    python_dict = {
        "int": int,
        "float": float,
        "bool": bool,
        "str": str,
        "bytes": bytes,
        "np.ndarray": np.ndarray,
        "image": Image,
        "depthimage": DepthImage,
        "camera": Camera,
        "compressedrle": CompressedRLE,
        "pose": Pose,
        "bbox": BBox,
        "gtinfo": GtInfo,
    }

    if isinstance(field, str):
        if field.startswith("[") and field.endswith("]"):
            return list
        if field.startswith("vector(") and field.endswith(")"):
            return np.ndarray
        return python_dict[field.lower()]
    return None


def field_to_pyarrow(field: str) -> pa.DataType:
    """Return PyArrpw type from string field

    Args:
        field (str): String field

    Returns:
        pa.DataType: PyArrow type
    """

    pyarrow_dict = {
        "int": pa.int64(),
        "float": pa.float32(),
        "bool": pa.bool_(),
        "str": pa.string(),
        "bytes": pa.binary(),
        "np.ndarray": pa.list_(pa.float32()),
        "image": ImageType,
        "depthimage": DepthImageType,
        "camera": CameraType,
        "compressedrle": CompressedRLEType,
        "pose": PoseType,
        "bbox": BBoxType,
        "gtinfo": GtInfoType,
    }

    if isinstance(field, str):
        if field.startswith("[") and field.endswith("]"):
            return pa.list_(
                pyarrow_dict[field.removeprefix("[").removesuffix("]").lower()]
            )
        if field.startswith("vector(") and field.endswith(")"):
            size_str = field.removeprefix("vector(").removesuffix(")")
            if size_str.isnumeric():
                return pa.list_(pa.float32(), list_size=int(size_str))
        return pyarrow_dict[field.lower()]
    return None


class Fields(BaseModel):
    """Dataset PyArrow fields as string dictionary

    Attributes:
        field_dict (dict[str, str]): PyArrow fields as string dictionary

    """

    field_dict: dict[str, str]

    def __init__(self, field_dict: dict[str, str]) -> None:
        """Create Fields from string dictionary

        Args:
            field_dict (dict[str, str]): PyArrow fields as string dictionary
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__(field_dict=field_dict)

    def to_schema(self) -> pa.schema:
        """Convert Fields string dictionary to PyArrow schema

        Returns:
            pa.schema: Fields as PyArrow schema
        """
        fields = []
        for field_name, field_type in self.field_dict.items():
            # Convert the field type to PyArrow type
            field = pa.field(field_name, field_to_pyarrow(field_type), nullable=True)
            fields.append(field)
        return pa.schema(fields)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import pyarrow as pa
from pydantic import BaseModel

from pixano.core.bbox import BBox, BBoxType
from pixano.core.pixano_type import PixanoType, create_pyarrow_type


class GtInfo(PixanoType, BaseModel):
    """GtInfo type

    Attributes:
        bbox_obj (BBox): bbox_obj
        bbox_visib (BBox): bbox_visib
        px_count_all (int): px_count_all
        px_count_valid (int): px_count_valid
        px_count_visib (int): px_count_visib
        visib_fract (float): visib_fract
    """

    bbox_obj: BBox
    bbox_visib: BBox
    px_count_all: int
    px_count_valid: int
    px_count_visib: int
    visib_fract: float

    def __init__(
        self,
        bbox_obj: BBox,
        bbox_visib: BBox,
        px_count_all: int,
        px_count_valid: int,
        px_count_visib: int,
        visib_fract: float,
    ):
        """Initialize GtInfo

        Args:
            bbox_obj (BBox): bbox_obj
            bbox_visib (BBox): bbox_visib
            px_count_all (int): px_count_all
            px_count_valid (int): px_count_valid
            px_count_visib (int): px_count_visib
            visib_fract (float): visib_fract
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__(
            bbox_obj=bbox_obj,
            bbox_visib=bbox_visib,
            px_count_all=px_count_all,
            px_count_valid=px_count_valid,
            px_count_visib=px_count_visib,
            visib_fract=visib_fract,
        )

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return GtInfo type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("bbox_obj", BBoxType),
                pa.field("bbox_visib", BBoxType),
                pa.field("px_count_all", pa.int64()),
                pa.field("px_count_valid", pa.int64()),
                pa.field("px_count_visib", pa.int64()),
                pa.field("visib_fract", pa.float64()),
            ]
        )


GtInfoType = create_pyarrow_type(GtInfo.to_struct(), "GtInfo", GtInfo)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from abc import ABC, abstractmethod
from pathlib import Path

from pixano.data.dataset import Dataset


class Exporter(ABC):
    """Abstract Data Exporter class

    Attributes:
        dataset (Dataset): Dataset to export
    """

    dataset: Dataset

    def __init__(
        self,
        input_dir: Path,
    ):
        """Initialize Exporter

        Args:
            input_dir (Path): Input dataset directory
        """

        # Dataset to export
        self.dataset = Dataset(input_dir)

    @abstractmethod
    def export_dataset(self, export_dir: Path):
        """Export dataset back to original format

        Args:
            export_dir (Path): Export directory
        """



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import random
import shutil
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Iterator
from datetime import timedelta
from io import BytesIO
from pathlib import Path

import lance
import lancedb
import pyarrow as pa
import shortuuid
from PIL import Image
from tqdm.auto import tqdm

from pixano.data.dataset import Dataset, DatasetInfo, DatasetTable
from pixano.data.fields import Fields
from pixano.data.item.item_feature import FeaturesValues
from pixano.utils import estimate_size


class Importer(ABC):
    """Dataset Importer class

    Attributes:
        info (DatasetInfo): Dataset information
        input_dirs (dict[str, Path]): Dataset input directories
    """

    info: DatasetInfo
    input_dirs: dict[str, Path]

    def __init__(
        self,
        name: str,
        description: str,
        tables: dict[str, list[DatasetTable]],
        splits: list[str],
        features_values: FeaturesValues = None,
    ):
        """Initialize Importer

        Args:
            name (str): Dataset name
            description (str): Dataset description
            tables (dict[str, list[DatasetTable]]): Dataset tables
            splits (list[str]): Dataset splits
            features_values (FeaturesValues, optional): Values for features
        """

        # Check input directories
        for source_path in self.input_dirs.values():
            if not source_path.exists():
                raise FileNotFoundError(f"{source_path} does not exist.")
            if not any(source_path.iterdir()):
                raise FileNotFoundError(f"{source_path} is empty.")

        # Create DatasetInfo
        self.info = DatasetInfo(
            id=shortuuid.uuid(),
            name=name,
            description=description,
            estimated_size="N/A",
            num_elements=0,
            splits=splits,
            tables=tables,
            features_values=features_values,
        )

    def create_tables(
        self, media_fields: dict[str, str] = None, object_fields: dict[str, str] = None
    ):
        """Create dataset tables

        Args:
            media_fields (dict[str, str], optional): Media fields. Defaults to None.
            object_fields (dict[str, str], optional): Object fields. Defaults to None.

        Returns:
            dict[str, list[DatasetTable]]: Tables
        """

        if media_fields is None:
            media_fields = {"image": "image"}

        tables: dict[str, list[DatasetTable]] = {
            "main": [
                DatasetTable(
                    name="db",
                    fields={
                        "id": "str",
                        "original_id": "str",
                        "views": "[str]",
                        "split": "str",
                    },
                )
            ],
            "media": [],
        }

        # Add media fields
        for field_name, field_type in media_fields.items():
            table_exists = False
            # If table for given field type exists
            for media_table in tables["media"]:
                if field_type == media_table.name and not table_exists:
                    media_table.fields[field_name] = field_type
                    table_exists = True
            # Else, create that table
            if not table_exists:
                tables["media"].append(
                    DatasetTable(
                        name=field_type,
                        fields={
                            "id": "str",
                            field_name: field_type,
                        },
                    )
                )

        # Add object fields
        if object_fields is not None:
            tables["objects"] = [
                DatasetTable(
                    name="objects",
                    fields={"id": "str", "item_id": "str", "view_id": "str"}
                    | object_fields,
                    source="Ground Truth",
                )
            ]

        return tables

    def create_preview(
        self,
        import_dir: Path,
        ds_tables: dict[str, dict[str, lancedb.db.LanceTable]],
    ):
        """Create dataset preview image

        Args:
            import_dir (Path): Import directory
            ds_tables (dict[str, dict[str, lancedb.db.LanceTable]]): Dataset tables
        """

        # Get list of image fields
        if "media" in ds_tables:
            if "image" in ds_tables["media"]:
                image_table = ds_tables["media"]["image"]
                if len(image_table) > 0:
                    image_fields = [
                        field.name for field in image_table.schema if field.name != "id"
                    ]
                    with tqdm(desc="Creating dataset thumbnail", total=1) as progress:
                        tile_w = 64
                        tile_h = 64
                        preview = Image.new("RGB", (4 * tile_w, 2 * tile_h))
                        for i in range(8):
                            field = image_fields[i % len(image_fields)]
                            item_id = random.randrange(len(image_table))
                            item = image_table.to_lance().take([item_id]).to_pylist()[0]
                            with Image.open(BytesIO(item[field].preview_bytes)) as im:
                                preview.paste(
                                    im,
                                    ((i % 4) * tile_w, (int(i / 4) % 2) * tile_h),
                                )
                        preview.save(import_dir / "preview.png")
                        progress.update(1)

    def copy_or_move_files(
        self,
        import_dir: Path,
        ds_tables: dict[str, dict[str, lancedb.db.LanceTable]],
        copy: bool,
    ):
        """Copy or move dataset files

        Args:
            import_dir (Path): Import directory
            ds_tables (dict[str, dict[str, lancedb.db.LanceTable]]): Dataset tables
            copy (bool): True to copy files, False to move them
        """

        if copy:
            for table in tqdm(
                ds_tables["media"].values(), desc="Copying media directories"
            ):
                for field in table.schema:
                    if field.name in self.input_dirs:
                        field_dir = import_dir / "media" / field.name
                        field_dir.mkdir(parents=True, exist_ok=True)
                        if self.input_dirs[field.name] != field_dir:
                            shutil.copytree(
                                self.input_dirs[field.name],
                                field_dir,
                                dirs_exist_ok=True,
                            )
        else:
            for table in tqdm(
                ds_tables["media"].values(), desc="Moving media directories"
            ):
                for field in table.schema:
                    if field.name in self.input_dirs:
                        field_dir = import_dir / "media" / field.name
                        if self.input_dirs[field.name] != field_dir:
                            self.input_dirs[field.name].rename(field_dir)

    @abstractmethod
    def import_rows(self) -> Iterator:
        """Process dataset rows for import

        Yields:
            Iterator: Processed rows
        """

    def import_dataset(
        self,
        import_dir: Path,
        copy: bool = True,
    ) -> Dataset:
        """Import dataset to Pixano format

        Args:
            import_dir (Path): Import directory
            copy (bool, optional): True to copy files to the import directory, False to move them. Defaults to True.

        Returns:
            Dataset: Imported dataset
        """

        # Create dataset
        dataset = Dataset.create(import_dir, self.info)

        # Load dataset tables
        ds_tables = dataset.open_tables()

        # Initalize batches
        ds_batches: dict[str, dict[str, list]] = defaultdict(dict)
        for group_name, table_group in self.info.tables.items():
            for table in table_group:
                ds_batches[group_name][table.name] = []

        # Add rows to tables
        save_batch_size = 1024
        for rows in tqdm(self.import_rows(), desc="Importing dataset"):
            for group_name, table_group in self.info.tables.items():
                for table in table_group:
                    # Store rows in a batch
                    ds_batches[group_name][table.name].extend(
                        rows[group_name][table.name]
                    )
                    # If batch reaches 1024 rows, store in table
                    if len(ds_batches[group_name][table.name]) >= save_batch_size:
                        pa_batch = pa.Table.from_pylist(
                            ds_batches[group_name][table.name],
                            schema=Fields(table.fields).to_schema(),
                        )
                        lance.write_dataset(
                            pa_batch,
                            uri=ds_tables[group_name][table.name].to_lance().uri,
                            mode="append",
                        )
                        ds_batches[group_name][table.name] = []

        # Store final batches
        for group_name, table_group in self.info.tables.items():
            for table in table_group:
                if len(ds_batches[group_name][table.name]) > 0:
                    pa_batch = pa.Table.from_pylist(
                        ds_batches[group_name][table.name],
                        schema=Fields(table.fields).to_schema(),
                    )
                    lance.write_dataset(
                        pa_batch,
                        uri=ds_tables[group_name][table.name].to_lance().uri,
                        mode="append",
                    )
                    ds_batches[group_name][table.name] = []

        # Optimize and clear creation history
        for tables in ds_tables.values():
            for table in tables.values():
                table.to_lance().optimize.compact_files()
                table.to_lance().cleanup_old_versions(older_than=timedelta(0))

        # Refresh tables
        ds_tables = dataset.open_tables()

        # Raise error if generated dataset is empty
        if len(ds_tables["main"]["db"]) == 0:
            raise FileNotFoundError(
                "Generated dataset is empty. Please make sure that the paths to your media files are correct, and that they each contain subfolders for your splits."
            )

        # Create DatasetInfo
        dataset.info.num_elements = len(ds_tables["main"]["db"])
        dataset.info.estimated_size = estimate_size(import_dir)
        dataset.save_info()

        # Create thumbnail
        self.create_preview(import_dir, ds_tables)

        # Copy or move media directories
        self.copy_or_move_files(import_dir, ds_tables, copy)

        return Dataset(import_dir)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Annotated

from fastapi import APIRouter, Depends, HTTPException

from pixano.data import Settings, get_settings

router = APIRouter(tags=["models"])


@router.get("/models", response_model=list[str])
async def get_models(
    settings: Annotated[Settings, Depends(get_settings)],
) -> list[str]:
    """Load models

    Returns:
        list[str]: List of models
    """

    # Load list of models
    models = []
    for model_path in settings.model_dir.glob("*.onnx"):
        models.append(model_path.name)

    # Return list of models
    if models:
        return models
    raise HTTPException(
        status_code=404,
        detail=f"No models found in {settings.model_dir.absolute()}",
    )



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import pyarrow as pa
from pydantic import BaseModel, PrivateAttr

from pixano.core.pixano_type import PixanoType, create_pyarrow_type


class Pose(PixanoType, BaseModel):
    """Pose type using orientation and translation matrices

    Attributes:
        _cam_r_m2c (list[float]): 3*3 orientation matrix
        _cam_t_m2c (list[float]): 3*1 translation matrix
    """

    _cam_r_m2c: list[float] = PrivateAttr()
    _cam_t_m2c: list[float] = PrivateAttr()

    def __init__(self, cam_r_m2c: list[float], cam_t_m2c: list[float]):
        """Initialize pose from orientation and translation matrices

        Args:
            cam_r_m2c (list[float]): 3*3 orientation matrix
            cam_t_m2c (list[float]): 3*1 translation matrix
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__()

        # Define private attributes manually
        self._cam_r_m2c = cam_r_m2c
        self._cam_t_m2c = cam_t_m2c

    @property
    def cam_r_m2c(self) -> list[float]:
        """Return Pose orientation matrix

        Returns:
            list[float]: 3*3 orientation matrix
        """

        return self._cam_r_m2c

    @property
    def cam_t_m2c(self) -> list[float]:
        """Return Pose translation matrix

        Returns:
            list[float]: 1*3 translation matrix
        """

        return self._cam_t_m2c

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return Pose type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("cam_r_m2c", pa.list_(pa.float64(), list_size=9)),
                pa.field("cam_t_m2c", pa.list_(pa.float64(), list_size=3)),
            ]
        )


PoseType = create_pyarrow_type(Pose.to_struct(), "Pose", Pose)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from abc import ABC
from datetime import datetime, timedelta
from math import ceil
from pathlib import Path

import duckdb
import lance
import lancedb
import pyarrow as pa
from tqdm.auto import tqdm

from pixano.data import Dataset, DatasetTable, Fields


class InferenceModel(ABC):
    """Abstract parent class for OfflineModel and OnlineModel

    Attributes:
        name (str): Model name
        model_id (str): Model ID
        device (str): Model GPU or CPU device
        description (str): Model description
    """

    def __init__(
        self,
        name: str,
        model_id: str = "",
        device: str = "",
        description: str = "",
    ) -> None:
        """Initialize model name and ID

        Args:
            name (str): Model name
            model_id (str, optional): Model ID. Defaults to "".
            device (str, optional): Model GPU or CPU device. Defaults to "".
            description (str, optional): Model description. Defaults to "".
        """

        self.name = name
        if model_id == "":
            self.model_id = f"{datetime.now().strftime('%y%m%d_%H%M%S')}_{name}"
        else:
            self.model_id = model_id
        self.device = device
        self.description = description

    def create_table(
        self,
        process_type: str,
        views: list[str],
        dataset: Dataset,
    ) -> DatasetTable:
        """Create inference table in dataset

        Args:
            process_type (str): Process type
                                - 'pre_ann' for pre-annotations to accept or reject as Ground Truth
                                - 'model_run' for annotations to compare to Ground Truth
                                - 'segment_emb' for segmentation embeddings
                                - 'search_emb' for semantic search embeddings
            views (list[str]): Dataset views
            dataset (Dataset): Dataset

        Returns:
            DatasetTable: Inference table
        """

        # Inference table filename
        table_filename = (
            f"emb_{self.model_id}" if "emb" in process_type else f"obj_{self.model_id}"
        )

        # Annotations schema
        if process_type in ["pre_ann", "model_run"]:
            table_group = "objects"
            # Create table
            table = DatasetTable(
                name=table_filename,
                fields={
                    "id": "str",
                    "item_id": "str",
                    "view_id": "str",
                    "bbox": "bbox",
                    "mask": "compressedrle",
                    "category": "str",
                },
                source=self.name if process_type == "model_run" else "Pre-annotation",
                type=None,
            )
        # Segmentation embeddings schema
        elif process_type == "segment_emb":
            table_group = "embeddings"
            # Add embedding column for each selected view
            fields = {"id": "str"}
            for view in views:
                fields[view] = "bytes"
            # Create table
            table = DatasetTable(
                name=table_filename,
                fields=fields,
                source=self.name,
                type="segment",
            )

        # Semantic search embeddings schema
        elif process_type == "search_emb":
            table_group = "embeddings"
            # Add vector column for each selected view
            fields = {"id": "str"}
            for view in views:
                fields[view] = "vector(512)"
            # Create table
            table = DatasetTable(
                name=table_filename,
                fields=fields,
                source=self.name,
                type="search",
            )

        # Create table
        dataset.create_table(table, table_group)

        return table

    def preannotate(
        self,
        batch: pa.RecordBatch,
        views: list[str],
        uri_prefix: str,
        threshold: float = 0.0,
        prompt: str = "",
    ) -> list[dict]:
        """Generate annotations for dataset rows

        Args:
            batch (pa.RecordBatch): Input batch
            views (list[str]): Dataset views
            uri_prefix (str): URI prefix for media files
            threshold (float, optional): Confidence threshold. Defaults to 0.0.
            prompt (str, optional): Annotation text prompt. Defaults to "".

        Returns:
            list[dict]: Annotation rows
        """

    def precompute_embeddings(
        self,
        batch: pa.RecordBatch,
        views: list[str],
        uri_prefix: str,
    ) -> list[dict]:
        """Precompute embeddings for dataset rows

        Args:
            batch (pa.RecordBatch): Input batch
            views (list[str]): Dataset views
            uri_prefix (str): URI prefix for media files

        Returns:
            list[dict]: Embedding rows
        """

    def process_dataset(
        self,
        dataset_dir: Path,
        views: list[str],
        process_type: str,
        splits: list[str] = None,
        batch_size: int = 1,
        threshold: float = 0.0,
        prompt: str = "",
    ) -> Dataset:
        """Process dataset for annotations or embeddings

        Args:
            dataset_dir (Path): Dataset directory
            views (list[str]): Dataset views
            process_type (str): Process type
                                - 'pre_ann' for pre-annotations to accept or reject as Ground Truth
                                - 'model_run' for annotations to compare to Ground Truth
                                - 'segment_emb' for segmentation embeddings
                                - 'search_emb' for semantic search embeddings
            splits (list[str], optional): Dataset splits, all if None. Defaults to None.
            batch_size (int, optional): Rows per process batch. Defaults to 1.
            threshold (float, optional): Confidence threshold for predictions. Defaults to 0.0.
            prompt (str, optional): Annotation text prompt. Defaults to "".

        Returns:
            Dataset: Dataset
        """

        if process_type not in [
            "pre_ann",
            "model_run",
            "segment_emb",
            "search_emb",
        ]:
            raise ValueError(
                "Please choose a valid process type"
                "('pre_ann' or 'model_run' for for annotations,"
                "'segment_emb' or 'search_emb' for segmentation or semantic search embeddings)"
            )

        if not views:
            raise ValueError("Please select which views you want to process on.")

        # Load dataset
        dataset = Dataset(dataset_dir)

        # Create inference table
        table = self.create_table(process_type, views, dataset)

        # Load dataset tables
        ds_tables = dataset.open_tables()

        # Load inference table
        table_group = "embeddings" if "emb" in process_type else "objects"
        table_lance = ds_tables[table_group][table.name].to_lance()

        # Create URI prefix
        uri_prefix = dataset.media_dir.absolute().as_uri()

        # Add rows to tables
        save_batch_size = 1024
        with tqdm(desc="Processing dataset", total=dataset.num_rows) as progress:
            for save_batch_index in range(ceil(dataset.num_rows / save_batch_size)):
                # Load rows
                process_batches = self._load_rows(
                    dataset,
                    ds_tables,
                    splits,
                    batch_size,
                    save_batch_size,
                    save_batch_index,
                )

                # Process rows
                save_batch = []
                for process_batch in process_batches:
                    save_batch.extend(
                        self.precompute_embeddings(process_batch, views, uri_prefix)
                        if "emb" in process_type
                        else self.preannotate(
                            process_batch, views, uri_prefix, threshold, prompt
                        )
                    )
                    progress.update(batch_size)

                # Save rows
                pyarrow_save_batch = pa.Table.from_pylist(
                    save_batch,
                    schema=Fields(table.fields).to_schema(),
                )
                lance.write_dataset(
                    pyarrow_save_batch,
                    uri=table_lance.uri,
                    mode="append",
                )

        # Optimize and clear creation history
        table_lance.optimize.compact_files()
        table_lance.cleanup_old_versions(older_than=timedelta(0))

        return dataset

    def _load_rows(
        self,
        dataset: Dataset,
        ds_tables: dict[str, dict[str, lancedb.db.LanceTable]],
        splits: list[str],
        process_batch_size: int,
        save_batch_size: int,
        save_batch_index: int,
    ) -> list[pa.RecordBatch]:
        """Load dataset rows as list of process batches

        Args:
            dataset (Dataset): Dataset
            ds_tables (dict[str, dict[str, lancedb.db.LanceTable]]): Dataset tables
            splits (list[str]): Dataset splits
            process_batch_size (int): Rows per process batch
            save_batch_size (int): Rows per save batch (batch of process batches)
            save_batch_index (int): Save batch index

        Returns:
            list[pa.RecordBatch]: Process batches
        """

        # Batch parameters
        offset = save_batch_index * save_batch_size
        limit = min(dataset.num_rows, offset + save_batch_size)

        # Main table
        # pylint: disable=unused-variable
        pyarrow_table = ds_tables["main"]["db"].to_lance()
        if splits:
            pyarrow_batch = duckdb.query(
                f"SELECT * FROM pyarrow_table WHERE split IN {tuple(splits)} ORDER BY len(id), id LIMIT {limit} OFFSET {offset}"
            ).to_arrow_table()
        else:
            pyarrow_batch = duckdb.query(
                f"SELECT * FROM pyarrow_table ORDER BY len(id), id LIMIT {limit} OFFSET {offset}"
            ).to_arrow_table()
        id_list = tuple(pyarrow_batch.to_pydict()["id"])

        # Media tables
        for media_table in ds_tables["media"].values():
            # pylint: disable=unused-variable
            pyarrow_media_table = media_table.to_lance()
            # pylint: disable=unused-variable
            pyarrow_media_batch = duckdb.query(
                f"SELECT * FROM pyarrow_media_table WHERE id in {id_list}"
            ).to_arrow_table()
            pyarrow_batch = duckdb.query(
                "SELECT * FROM pyarrow_batch LEFT JOIN pyarrow_media_batch USING (id) ORDER BY len(id), id"
            ).to_arrow_table()

        # Convert to RecordBatch
        return pyarrow_batch.to_batches(max_chunksize=process_batch_size)

    def export_to_onnx(self, library_dir: Path):
        """Export Torch model to ONNX

        Args:
            library_dir (Path): Dataset library directory
        """



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from fractions import Fraction
from typing import Any

import numpy as np
import pandas as pd
import pyarrow as pa


def compute_additional_data(data_table: pa.Table) -> pd.DataFrame:
    """Convert Table to DataFrame and add resolution and aspect ratio

    Args:
        data_table (pa.Table): Input Table

    Returns:
        pd.DataFrame: DataFrame with added resolution and aspect ratio
    """

    # Take a subset of table without image columns (which can't be converted to pandas)
    if not all(p in data_table.column_names for p in ["width", "height"]):
        return None
    data = data_table.select(["width", "height"]).to_pandas()

    # Compute additional data
    data["resolution"] = data.apply(
        lambda x: str(x["width"]) + "x" + str(x["height"]), axis=1
    )
    data["aspect_ratio"] = data.apply(
        lambda x: str(Fraction(x["width"], x["height"])).replace("/", ":"), axis=1
    )

    return data


def objects_table_to_df(data_table: pa.Table, field: str) -> pd.DataFrame:
    """Convert a field from the objects column to a DataFrame

    Args:
        data_table (pa.Table): Table with an objects column
        field (str): Selected field from the objects column

    Returns:
        pd.DataFrame: Selected field as DataFrame
    """

    try:
        df_objs = data_table.select(["objects"]).to_pandas()
        sel = [{field: d[field]} for objs in df_objs["objects"] for d in objs]
        return pd.DataFrame.from_dict(sel)
    except ValueError as e:
        raise ValueError("Unable to convert table Pandas DataFrame") from e


def categorical_stats(df: pd.DataFrame, split: str, field_name: str) -> list[dict]:
    """Compute feature categorical statistics

    Args:
        df (pd.DataFrame): Input DataFrame
        split (str): DataFrame split
        field_name (str): Selected field

    Returns:
        list[dict]: Feature statistics
    """

    counts = df.value_counts(subset=field_name)
    return [{field_name: k, "counts": v, "split": split} for k, v in counts.items()]


def numerical_stats(
    df: pd.DataFrame, split: str, field_name: str, field_range: list[float] = None
) -> list[dict]:
    """Compute feature numerical statistics

    Args:
        df (pd.DataFrame): Input DataFrame
        split (str): DataFrame split
        field_name (str): Selected field
        field_range (list[float], optional): Selected field range. Defaults to None.

    Returns:
        list[dict]: Feature statistics
    """

    counts, bins = np.histogram(df[field_name], range=field_range)
    return [
        {
            "bin_start": float(bins[i]),
            "bin_end": float(bins[i + 1]),
            "counts": int(counts[i]),
            "split": split,
        }
        for i in range(len(counts))
    ]


def compute_stats(df: pd.DataFrame, split: str, feature: dict[str, Any]) -> list[dict]:
    """Compute feature statistics

    Args:
        df (pd.DataFrame): Input DataFrame
        split (str): DataFrame split
        feature (dict): Selected feature

    Returns:
        list[dict]: Feature statistics
    """

    # Categorical
    if feature["type"] == "categorical":
        return categorical_stats(df, split, feature["name"])
    # Numerical
    if feature["type"] == "numerical":
        return numerical_stats(df, split, feature["name"], feature.get("range", None))
    return []



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from functools import lru_cache
from pathlib import Path
from typing import Optional
from urllib.parse import urlparse

import boto3
from pydantic import ConfigDict
from pydantic_settings import BaseSettings
from s3path import S3Path, register_configuration_parameter


class Settings(BaseSettings):
    """Pixano app settings

    Attributes:
        library_dir (str): Local or S3 path to dataset library
        aws_endpoint (str): S3 endpoint URL, use 'AWS' if not provided. Used if library_dir is an S3 path
        aws_region (str): S3 region name, not always required for private storages. Used if library_dir is an S3 path
        aws_access_key (str): S3 AWS access key. Used if library_dir is an S3 path
        aws_secret_key (str): S3 AWS secret key. Used if library_dir is an S3 path
        local_model_dir (str): Local path to models. Used if library_dir is an S3 path
        data_dir (Path | S3Path): Dataset directory as Path | S3Path
        model_dir (Path): Model directory as Path
    """

    library_dir: Optional[str] = (Path.cwd() / "library").as_posix()
    aws_endpoint: Optional[str] = None
    aws_region: Optional[str] = None
    aws_access_key: Optional[str] = None
    aws_secret_key: Optional[str] = None
    local_model_dir: Optional[str] = None
    data_dir: Optional[Path | S3Path] = None
    model_dir: Optional[Path] = None
    # Change Pydantic protected namespace from "model_" to "settings_" because of model_dir
    model_config = ConfigDict(protected_namespaces=("settings_",))

    def __init__(self, *args, **kwargs):
        """Initialize settings"""

        super().__init__(*args, **kwargs)

        # Setup data directory
        if urlparse(self.library_dir).scheme == "s3":
            # S3 library
            try:
                self.data_dir = S3Path.from_uri(self.library_dir)
                register_configuration_parameter(
                    self.data_dir,
                    resource=boto3.resource(
                        "s3",
                        endpoint_url=self.aws_endpoint,
                        region_name=self.aws_region,
                        aws_access_key_id=self.aws_access_key,
                        aws_secret_access_key=self.aws_secret_key,
                    ),
                )
            except Exception as e:
                raise ValueError(
                    "ERROR: Could not register S3 dataset library.\n"
                    "You have to set the following environment variables:\n"
                    "- AWS_ENDPOINT: S3 endpoint URL, use 'AWS' if not provided\n"
                    "- AWS_REGION: S3 region name, not always required for private storages\n"
                    "- AWS_ACCESS_KEY: S3 AWS access key\n"
                    "- AWS_SECRET_KEY: S3 AWS secret key"
                ) from e

            # Check if local model directory is provided
            if self.local_model_dir is None:
                raise AttributeError(
                    "When using S3 storage, runtime models (.onnx files) must be stored locally and their directory must be provided with LOCAL_MODEL_DIR."
                )
            self.model_dir = Path(self.local_model_dir)

        else:
            # Local library
            self.data_dir = Path(self.library_dir)
            self.model_dir = self.data_dir / "models"


@lru_cache
def get_settings() -> Settings:
    """Get app settings

    Returns:
        Settings: App settings
    """

    return Settings()



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Optional

import numpy as np
import pyarrow as pa
from pydantic import BaseModel, PrivateAttr

from pixano.core.compressed_rle import CompressedRLE
from pixano.core.pixano_type import PixanoType, create_pyarrow_type
from pixano.utils import (
    denormalize_coords,
    mask_to_bbox,
    normalize_coords,
    xywh_to_xyxy,
    xyxy_to_xywh,
)

# Disable warning for BBox "format" attribute
# NOTE: Rename attribute? Breaking change for Pixano datasets using BBox
# pylint: disable=redefined-builtin


class BBox(PixanoType, BaseModel):
    """Bounding box type using coordinates in xyxy or xywh format

    Attributes:
        coords (list[float]): List of coordinates in given format
        format (str): Coordinates format, 'xyxy' or 'xywh'
        is_normalized (bool, optional): True if coordinates are normalized to image size
        confidence (float, optional): Bounding box confidence if predicted
    """

    _coords: list[float] = PrivateAttr()
    _format: str = PrivateAttr()
    _is_normalized: Optional[bool] = PrivateAttr()
    _confidence: Optional[float] = PrivateAttr()

    def __init__(
        self,
        coords: list[float],
        format: str,
        is_normalized: bool = True,
        confidence: float = None,
    ):
        """Initialize Bounding box

        Args:
            coords (list[float]): List of coordinates in given format
            format (str): Coordinates format, 'xyxy' or 'xywh'
            is_normalized (bool, optional): True if coordinates are normalized to image size. Defaults to True.
            confidence (float, optional): Bounding box confidence if predicted. Defaults to None.
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__()

        # Define private attributes manually
        self._coords = coords
        self._format = format
        self._is_normalized = is_normalized
        self._confidence = confidence

    @property
    def coords(self) -> list[float]:
        """Return bounding box coordinates

        Returns:
            list[float]: Coordinates
        """

        return self._coords

    @property
    def xyxy_coords(self) -> list[float]:
        """Return bounding box xyxy coordinates

        Returns:
            list[float]: Coordinates in xyxy format
        """

        return self.coords if self.format == "xyxy" else xywh_to_xyxy(self.coords)

    @property
    def xywh_coords(self) -> list[float]:
        """Return bounding box xywh coordinates

        Returns:
            list[float]: Coordinates in xywh format
        """

        return self.coords if self.format == "xywh" else xyxy_to_xywh(self.coords)

    @property
    def format(self) -> str:
        """Return bounding box coordinates format

        Returns:
            str: Coordinates format, 'xyxy' or 'xywh'
        """

        return self._format

    @property
    def is_normalized(self) -> bool:
        """Return bounding box normalization information

        Returns:
            bool: True if coordinates are normalized to image size
        """

        return self._is_normalized

    @property
    def confidence(self) -> float:
        """Return bounding box confidence

        Returns:
            float: Bounding box confidence if predicted, else None
        """

        return self._confidence

    @property
    def is_predicted(self) -> bool:
        """Return True if bounding box is predicted and has a confidence value

        Returns:
            bool: True if bounding box is predicted and has a confidence value
        """

        return self.confidence is not None

    def to_xyxy(self) -> "BBox":
        """Return bounding box in xyxy format

        Returns:
            BBox: Bounding box in xyxy format
        """

        return BBox(self.xyxy_coords, "xyxy", self.is_normalized, self.confidence)

    def to_xywh(self) -> "BBox":
        """Return bounding box in xywh format

        Returns:
            BBox: Bounding box in xyxy format
        """

        return BBox(self.xywh_coords, "xywh", self.is_normalized, self.confidence)

    def normalize(self, height: int, width: int) -> "BBox":
        """Return bounding box with coordinates normalized to image size

        Args:
            height (int): Image height
            width (int): Image width

        Returns:
            BBox: Bounding box with coordinates normalized to image size
        """

        return BBox(
            normalize_coords(self.coords, height, width),
            self.format,
            True,
            self.confidence,
        )

    def denormalize(self, height: int, width: int) -> "BBox":
        """Return bounding box with coordinates denormalized from image size

        Args:
            height (int): Image height
            width (int): Image width

        Returns:
            BBox: Bounding box with coordinates denormalized from image size
        """

        return BBox(
            denormalize_coords(self.coords, height, width),
            self.format,
            False,
            self.confidence,
        )

    @staticmethod
    def from_xyxy(xyxy: list[float], confidence: float = None) -> "BBox":
        """Create bounding box using normalized xyxy coordinates

        Args:
            xyxy (list[float]): List of coordinates in xyxy format
            confidence (float, optional): Bounding box confidence if predicted. Defaults to None.

        Returns:
            Bbox: Bounding box
        """

        return BBox(xyxy, "xyxy", confidence=confidence)

    @staticmethod
    def from_xywh(xywh: list[float], confidence: float = None) -> "BBox":
        """Create bounding box using normalized xywh coordinates

        Args:
            xywh (list[float]): List of coordinates in xywh format
            confidence (float, optional): Bounding box confidence if predicted. Defaults to None.

        Returns:
            Bbox: Bounding box
        """

        return BBox(xywh, "xywh", confidence=confidence)

    @staticmethod
    def from_mask(mask: np.ndarray) -> "BBox":
        """Create bounding box using a NumPy array mask

        Args:
            mask (np.ndarray): NumPy array mask

        Returns:
            Bbox: Bounding box
        """

        return BBox.from_xywh(mask_to_bbox(mask))

    @staticmethod
    def from_rle(rle: CompressedRLE) -> "BBox":
        """Create bounding box using a RLE mask

        Args:
            rle (CompressedRLE): RLE mask

        Returns:
            Bbox: Bounding box
        """

        return BBox.from_mask(rle.to_mask())

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return BBox type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("coords", pa.list_(pa.float32(), list_size=4)),
                pa.field("is_normalized", pa.bool_()),
                pa.field("format", pa.string()),
                pa.field("confidence", pa.float32()),
            ]
        )


BBoxType = create_pyarrow_type(BBox.to_struct(), "Bbox", BBox)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import io
from typing import IO, Optional

import imageio
import matplotlib.pyplot as plt
import numpy as np
import pyarrow as pa
from pydantic import BaseModel, PrivateAttr

from pixano.core.pixano_type import PixanoType, create_pyarrow_type

# Disable warning for DepthImage "bytes" attribute
# NOTE: Rename attribute? Breaking change for Pixano datasets using DepthImage
# pylint: disable=redefined-builtin, used-before-assignment


class DepthImage(PixanoType, BaseModel):
    """Depth image type

    Attributes:
        _depth_map (np.ndarray, optional): Depth image as NumPy array
        _bytes (bytes, optional): Depth image as bytes
        _shape (list[int], optional): Depth image shape
    """

    _depth_map: Optional[np.ndarray] = PrivateAttr()
    _bytes: Optional[bytes] = PrivateAttr()
    _shape: Optional[list[int]] = PrivateAttr()

    def __init__(
        self,
        depth_map: np.ndarray = None,
        bytes: bytes = None,
        shape: list[int] = None,
    ):
        """Initialize Depth image

        Args:
            depth_map (np.ndarray, optional): Depth image as NumPy array. Defaults to None.
            bytes (bytes, optional): Depth image as bytes. Defaults to None.
            shape (list[int], optional): Depth image shape. Defaults to None.
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__()

        # Define private attributes manually
        self._depth_map = depth_map
        self._bytes = bytes
        self._shape = shape

    @property
    def shape(self) -> list[int]:
        """Return Depth image shape

        Returns:
            list[int]: Depth image shape
        """

        return self._shape

    @property
    def bytes(self) -> bytes:
        """Return Depth image as bytes

        Returns:
            bytes: Depth image as bytes
        """

        if self._bytes is not None:
            return self._bytes
        return self._depth_map.tobytes()

    @property
    def depth_map(self) -> np.ndarray:
        """Returns Depth image as NumPy array

        Returns:
            np.ndarray: Depth image as NumPy array
        """

        if self._depth_map is not None:
            return self._depth_map
        return np.frombuffer(self._bytes, dtype=np.float32).reshape(self._shape)

    @staticmethod
    def load_npy(path: str) -> "DepthImage":
        """Create depth image from .npy file

        Args:
            path (str): Path to .npy file containing depth image as NumPy Array.

        Returns:
            DepthImage: Depth image
        """

        depth_map = np.load(path)
        return DepthImage(depth_map=depth_map, shape=depth_map.shape)

    @staticmethod
    def load(path: str) -> "DepthImage":
        """Create depth image from 16-bit .png file

        Args:
            path (str): Path of .png file of depth image

        Returns:
            DepthImage: Depth image
        """

        depth_map = imageio.v3.imread(path).astype(np.uint16)
        return DepthImage(depth_map=depth_map, shape=depth_map.shape)

    def save(self, path):
        """Save depth image to .png file.

        Args:
            path (str): Path to .png file to save
        """

        depth_image = self.depth_map.astype(np.uint16)
        imageio.v3.imwrite(path, depth_image)

    def open(self) -> IO:
        """Open depth image

        Returns:
            IO: Opened depth image
        """

        return io.BytesIO(self.bytes)

    def to_grayscale(
        self,
    ) -> "DepthImage":
        """Transform Depth image to 8-bit grayscale depth image

        Returns:
            DepthImage: 8-bit grayscale depth image
        """

        depth = self.depth_map
        d_min, d_max = depth.min(), depth.max()
        depth_n: np.ndarray = ((depth - d_min) / (d_max - d_min)) * 255
        return DepthImage(depth_map=depth_n.astype(np.uint8), shape=depth.shape)

    def display(self):
        """Display Depth image with matplotlib

        Returns:
            plt.Figure: Plotted image
        """

        fig, ax = plt.subplots(figsize=self._shape)
        ax.imshow(self.depth_map.astype(np.int8), cmap="gray", vmin=0, vmax=255)
        ax.axis("off")

        plt.ion()
        plt.show()

        return fig

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return DepthImage type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("bytes", pa.binary()),
                pa.field("shape", pa.list_(pa.int32(), list_size=2)),
            ]
        )


DepthImageType = create_pyarrow_type(DepthImage.to_struct(), "DepthImage", DepthImage)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import glob
from collections.abc import Iterator
from pathlib import Path

import shortuuid
from PIL import Image as PILImage

from pixano.core import BBox, Image
from pixano.data.importers.importer import Importer
from pixano.data.item.item_feature import FeaturesValues, FeatureValues
from pixano.utils import image_to_thumbnail, natural_key


class DOTAImporter(Importer):
    """Importer class for DOTA dataset

    Attributes:
        info (DatasetInfo): Dataset information
        input_dirs (dict[str, Path]): Dataset input directories
    """

    def __init__(
        self,
        name: str,
        description: str,
        input_dirs: dict[str, Path],
        splits: list[str],
    ):
        """Initialize DOTA Importer

        Args:
            name (str): Dataset name
            description (str): Dataset description
            input_dirs (dict[str, Path]): Dataset input directories
            splits (list[str]): Dataset splits
        """

        # Create tables
        tables = super().create_tables(
            media_fields={"image": "image"},
            object_fields={
                "original_id": "str",
                "bbox": "bbox",
                "category": "str",
            },
        )

        # Create categories
        features_values = FeaturesValues(
            objects={
                "category": FeatureValues(
                    restricted=False,
                    values=[
                        "plane",  # id=1
                        "ship",  # id=2
                        "storage tank",  # id=3
                        "baseball diamond",  # id=4
                        "tennis court",  # id=5
                        "basketball court",  # id=6
                        "ground track field",  # id=7
                        "harbor",  # id=8
                        "bridge",  # id=9
                        "large vehicle",  # id=10
                        "small vehicle",  # id=11
                        "helicopter",  # id=12
                        "roundabout",  # id=13
                        "soccer ball field",  # id=14
                        "swimming pool",  # id=15
                        "container crane",  # id=16
                        "airport",  # id=17
                        "helipad",  # id=18
                    ],
                )
            }
        )

        # Initialize Importer
        self.input_dirs = input_dirs
        super().__init__(name, description, tables, splits, features_values)

    def import_rows(self) -> Iterator:
        """Process dataset rows for import

        Yields:
            Iterator: Processed rows
        """
        for split in self.info.splits:
            # Get images paths
            image_paths = glob.glob(str(self.input_dirs["image"] / split / "*.png"))
            image_paths = [Path(p) for p in sorted(image_paths, key=natural_key)]

            # Process rows
            for im_path in image_paths:
                # Load image annotations
                im_anns_file = (
                    self.input_dirs["objects"]
                    / split
                    / "hbb"
                    / im_path.name.replace("png", "txt")
                )
                with open(im_anns_file, encoding="utf-8") as f:
                    im_anns = [line.strip().split() for line in f]

                # Allow DOTA largest images
                PILImage.MAX_IMAGE_PIXELS = 806504000

                # Get image dimensions and thumbnail
                with PILImage.open(im_path) as im:
                    im_w, im_h = im.size
                    im_thumb = image_to_thumbnail(im)

                # Set image URI
                im_uri = f"image/{split}/{im_path.name}"

                # Set unique item id
                item_id = shortuuid.uuid()

                # Return rows
                rows = {
                    "main": {
                        "db": [
                            {
                                "id": item_id,
                                "original_id": im_path.stem,
                                "views": ["image"],
                                "split": split,
                            }
                        ]
                    },
                    "media": {
                        "image": [
                            {
                                "id": item_id,
                                "image": Image(im_uri, None, im_thumb).to_dict(),
                            }
                        ]
                    },
                    "objects": {
                        "objects": [
                            {
                                "id": shortuuid.uuid(),
                                "item_id": item_id,
                                "view_id": "image",
                                "bbox": BBox.from_xyxy(
                                    [
                                        float(ann[0]),
                                        float(ann[1]),
                                        float(ann[4]),
                                        float(ann[5]),
                                    ]
                                )
                                .normalize(im_h, im_w)
                                .to_dict(),
                                "category": str(ann[8]).replace("-", " "),
                            }
                            for ann in im_anns
                        ]
                    },
                }

                yield rows



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import base64

import pyarrow as pa
from pydantic import BaseModel

from pixano.core import is_binary


class ItemEmbedding(BaseModel):
    """Embedding type for DatasetItem

    Attributes:
        view_id (str): Embedding view ID
        data (str): Embedding data in base 64
    """

    view_id: str
    data: str

    @staticmethod
    def from_pyarrow(table: pa.Table, schema: pa.schema) -> dict[str, "ItemEmbedding"]:
        """Create dictionary of ItemEmbedding from PyArrow Table

        Args:
            table (dict[str, Any]): PyArrow table
            schema (pa.schema): PyArrow schema

        Returns:
            dict[str, ItemEmbedding]: Dictionary of ItemEmbedding
        """

        list_item = table.to_pylist()
        if len(list_item) == 0:
            return {}
        item = list_item[0]
        embeddings = {}

        # Iterate on fields
        for field in schema:
            # Image
            if is_binary(field.type):
                embeddings[field.name] = ItemEmbedding(
                    view_id=field.name,
                    data=base64.b64encode(item[field.name]).decode("ascii"),
                )

        return embeddings



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pixano.utils.boxes import (
    denormalize_coords,
    mask_to_bbox,
    normalize_coords,
    urle_to_bbox,
    xywh_to_xyxy,
    xyxy_to_xywh,
)
from pixano.utils.image import (
    binary_to_url,
    depth_array_to_gray,
    depth_file_to_binary,
    encode_rle,
    image_to_binary,
    image_to_thumbnail,
    mask_to_polygons,
    mask_to_rle,
    polygons_to_rle,
    rle_to_mask,
    rle_to_polygons,
    rle_to_urle,
    urle_to_rle,
)
from pixano.utils.labels import (
    coco_ids_80to91,
    coco_names_80,
    coco_names_91,
    dota_ids,
    voc_names,
)
from pixano.utils.python import estimate_size, natural_key

__all__ = [
    "normalize_coords",
    "denormalize_coords",
    "mask_to_bbox",
    "urle_to_bbox",
    "xywh_to_xyxy",
    "xyxy_to_xywh",
    "image_to_binary",
    "image_to_thumbnail",
    "binary_to_url",
    "depth_array_to_gray",
    "depth_file_to_binary",
    "encode_rle",
    "mask_to_rle",
    "rle_to_mask",
    "polygons_to_rle",
    "rle_to_polygons",
    "mask_to_polygons",
    "urle_to_rle",
    "rle_to_urle",
    "coco_ids_80to91",
    "coco_names_80",
    "coco_names_91",
    "dota_ids",
    "voc_names",
    "estimate_size",
    "natural_key",
]



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from collections import defaultdict
from pathlib import Path
from typing import Optional

import duckdb
import lancedb
import pyarrow as pa
import pyarrow.dataset as pa_ds
from pydantic import BaseModel, ConfigDict
from s3path import S3Path

from pixano.core import Image
from pixano.data.dataset.dataset_info import DatasetInfo
from pixano.data.dataset.dataset_item import DatasetItem
from pixano.data.dataset.dataset_stat import DatasetStat
from pixano.data.dataset.dataset_table import DatasetTable
from pixano.data.fields import Fields, field_to_pyarrow
from pixano.data.item.item_feature import FeaturesValues, FeatureValues, ItemFeature
from pixano.data.item.item_object import ItemObject


class Dataset(BaseModel):
    """Dataset

    Attributes:
        path (Path | S3Path): Dataset path
        info (DatasetInfo, optional): Dataset info
        stats (list[DatasetStat], optional): Dataset stats
        thumbnail (str, optional): Dataset thumbnail base 64 URL
    """

    path: Path | S3Path
    info: Optional[DatasetInfo] = None
    stats: Optional[list[DatasetStat]] = None
    thumbnail: Optional[str] = None
    # Allow arbitrary types because of S3 Path
    model_config = ConfigDict(arbitrary_types_allowed=True)

    def __init__(
        self,
        path: Path | S3Path,
    ):
        """Initialize dataset

        Args:
            path (Path | S3Path): Dataset path
        """

        info_file = path / "db.json"
        stats_file = path / "stats.json"
        thumb_file = path / "preview.png"

        # Define public attributes through Pydantic BaseModel
        super().__init__(
            path=path,
            info=DatasetInfo.from_json(info_file),
            stats=DatasetStat.from_json(stats_file) if stats_file.is_file() else None,
            thumbnail=(
                Image(uri=thumb_file.absolute().as_uri()).url
                if thumb_file.is_file()
                else None
            ),
        )

    @property
    def media_dir(self) -> Path | S3Path:
        """Return dataset media directory

        Returns:
            Path | S3Path: Dataset media directory
        """

        return self.path / "media"

    @property
    def num_rows(self) -> int:
        """Return number of rows in dataset

        Returns:
            int: Number of rows
        """

        ds = self.connect()

        # Return number of rows of main table
        return len(ds.open_table("db"))

    def get_item_uuid(self, original_id: str) -> str:
        """Get id (uuid) from original data item id, if exist

        Args:
            original_id (str): original data item id

        Returns:
            str: unique item id, or original_id if not found
        """
        ds_tables = self.open_tables()
        # pylint: disable=unused-variable
        lance_table = ds_tables["main"]["db"].to_lance()
        result = (
            duckdb.query(
                f"SELECT id FROM lance_table WHERE original_id = '{original_id}'"
            )
            .to_arrow_table()
            .to_pylist()
        )
        if result and len(result) == 1:
            return result[0]["id"]
        return original_id

    def get_object_uuid(self, original_id: str) -> str:
        """Get id (uuid) from original data object id, if exist

        Args:
            original_id (str): original data object id

        Returns:
            str: unique item id, or original_id if not found
        """
        ds_tables = self.open_tables()
        # pylint: disable=unused-variable
        lance_table = ds_tables["objects"]["objects"].to_lance()
        result = (
            duckdb.query(
                f"SELECT id FROM lance_table WHERE original_id = '{original_id}'"
            )
            .to_arrow_table()
            .to_pylist()
        )
        if result and len(result) == 1:
            return result[0]["id"]
        return original_id

    def load_info(
        self,
        load_stats: bool = False,
        load_thumbnail: bool = False,
        load_features_values: bool = False,
    ) -> DatasetInfo:
        """Return dataset info with thumbnail and stats inside

        Args:
            load_stats (bool, optional): Load dataset stats. Defaults to False.
            load_thumbnail (bool, optional): Load dataset thumbnail. Defaults to False.
            load_features_values (bool, optional): Load available values. Defaults to False.

        Returns:
            DatasetInfo: Dataset info
        """

        info = DatasetInfo.from_json(
            self.path / "db.json",
            load_stats=load_stats,
            load_thumbnail=load_thumbnail,
        )
        if load_features_values:
            info.features_values = self.get_features_values(info.features_values)

        return info

    def save_info(self):
        """Save updated dataset info"""

        self.info.save(self.path)

    @staticmethod
    def create(path: Path | S3Path, info: DatasetInfo) -> "Dataset":
        """Create dataset

        Args:
            path (Path | S3Path): Path to create dataset in
            info (DatasetInfo): Dataset info

        Returns:
            Dataset: Created dataset
        """

        # Create DatasetInfo file
        path.mkdir(parents=True, exist_ok=True)
        info.save(path)

        # Load dataset
        dataset = Dataset(path)

        # Create dataset tables
        for group_name, table_group in dataset.info.tables.items():
            for table in table_group:
                dataset.create_table(table, group_name, add_to_info=False)

        return dataset

    def connect(self) -> lancedb.db.DBConnection:
        """Connect to dataset with LanceDB

        Returns:
            lancedb.db.DBConnection: Dataset LanceDB connection
        """

        if isinstance(self.path, S3Path):
            return lancedb.connect(self.path.as_uri())
        return lancedb.connect(self.path)

    def open_tables(self) -> dict[str, dict[str, lancedb.db.LanceTable]]:
        """Open dataset tables with LanceDB

        Returns:
            dict[str, dict[str, lancedb.db.LanceTable]]: Dataset tables
        """

        ds = self.connect()

        ds_tables: dict[str, dict[str, lancedb.db.LanceTable]] = defaultdict(dict)

        for group_name, table_group in self.info.tables.items():
            for table in table_group:
                try:
                    ds_tables[group_name][table.name] = ds.open_table(table.name)
                except FileNotFoundError as e:
                    # If optional table, remove from DatasetInfo
                    if group_name in ["objects", "embeddings", "active_learning"]:
                        self.info.tables[group_name].remove(table)
                        self.save_info()
                    else:
                        raise FileNotFoundError from e

        return ds_tables

    def create_table(
        self,
        table: DatasetTable,
        table_group: str,
        add_to_info: bool = True,
    ):
        """Create a new table in the dataset

        Args:
            table (DatasetTable): Table to create
            table_group (str): Table group
            add_to_info (bool, optional): Add table to DatasetInfo. Defaults to True.
        """

        # Create Lance table
        ds = self.connect()
        # pylint: disable=unexpected-keyword-arg
        ds.create_table(
            table.name,
            schema=Fields(table.fields).to_schema(),
            mode="overwrite",
        )

        # Save table to DatasetInfo
        if add_to_info:
            if table_group in self.info.tables:
                self.info.tables[table_group].append(table)
            else:
                self.info.tables[table_group] = [table]
            self.save_info()

    def update_table(
        self,
        element: DatasetItem | ItemObject,
        table: lancedb.db.LanceTable,
        table_group: str,
        table_name: str,
    ):
        """Update a table with new features or base fields

        Args:
            element (DatasetItem | ItemObject): Table element (item or object)
            table (lancedb.db.LanceTable): Table to update
            table_group (str): Table group
            table_name (str): Table name
        """

        new_columns: list[ItemFeature] = []

        # Check for new base fields
        base_fields = {"review_state": "str", "bbox": "bbox", "mask": "compressedrle"}
        new_columns.extend(
            [
                ItemFeature(name=field_name, dtype=field_type, value=None)
                for field_name, field_type in base_fields.items()
                if hasattr(element, field_name) and field_name not in table.schema.names
            ]
        )
        # Check for new features
        if element.features is not None:
            new_columns.extend(
                [
                    feat
                    for feat in element.features.values()
                    if feat.name not in table.schema.names
                ]
            )
        # Add new columns
        if len(new_columns) > 0:
            new_columns_table = table.to_lance().to_table(columns=["id"])
            for col in new_columns:
                # None should be suported for booleans with pylance 0.9.1
                # None is not supported for integers and floats yet
                none_value = (
                    False
                    if col.dtype == "bool"
                    else 0 if col.dtype in ("int", "float") else None
                )
                col_array = pa.array(
                    [none_value] * len(table),
                    type=field_to_pyarrow(col.dtype),
                )
                new_columns_table = new_columns_table.append_column(
                    pa.field(col.name, field_to_pyarrow(col.dtype)), col_array
                )
                # Update DatasetInfo
                for info_table in self.info.tables[table_group]:
                    if info_table.name == table_name:
                        info_table.fields[col.name] = col.dtype

            # Merge with main table
            table.to_lance().merge(new_columns_table, "id")
            self.save_info()

    def load_items(
        self,
        limit: int,
        offset: int,
        load_active_learning: bool = True,
    ) -> list[DatasetItem]:
        """Load dataset items in selected tables

        Args:
            limit (int): Items limit
            offset (int): Items offset
            load_active_learning (bool, optional): Load items active learning info. Defaults to True.
        Returns:
            list[DatasetItem]: List of dataset items
        """

        # Update info in case of change
        self.info = self.load_info()

        # Load tables
        ds_tables = self.open_tables()

        # Load PyArrow items from tables
        pyarrow_items: dict[str, dict[str, pa.Table]] = defaultdict(dict)

        # Load PyArrow items from main table
        # pylint: disable=unused-variable
        lance_table = ds_tables["main"]["db"].to_lance()
        id_field = "original_id" if "original_id" in lance_table.schema.names else "id"
        pyarrow_items["main"]["db"] = duckdb.query(
            f"SELECT * FROM lance_table ORDER BY len({id_field}), {id_field} LIMIT {limit} OFFSET {offset}"
        ).to_arrow_table()
        id_list = tuple(pyarrow_items["main"]["db"].to_pydict()["id"])

        # Media tables
        for media_source, media_table in ds_tables["media"].items():
            lance_table = media_table.to_lance()
            pyarrow_items["media"][media_source] = duckdb.query(
                f"SELECT * FROM lance_table WHERE id IN {id_list}"
            ).to_arrow_table()

        # Active Learning tables
        if load_active_learning:
            for al_source, al_table in ds_tables["active_learning"].items():
                lance_table = al_table.to_lance()
                pyarrow_items["active_learning"][al_source] = duckdb.query(
                    f"SELECT * FROM lance_table  WHERE id IN {id_list}"
                ).to_arrow_table()

        if pyarrow_items["main"]["db"].num_rows > 0:
            # Split results
            pyarrow_item_list = self._split_items(pyarrow_items, load_active_learning)

            return [
                DatasetItem.from_pyarrow(pyarrow_item, self.info, self.media_dir)
                for pyarrow_item in pyarrow_item_list
            ]
        return None

    def search_items(
        self,
        limit: int,
        offset: int,
        query: dict[str, str],
        load_active_learning: bool = True,
    ):
        """Search for dataset items in selected tables

        Args:
            limit (int): Items limit
            offset (int): Items offset
            query (dict[str, str]): Search query
            load_active_learning (bool, optional): Load items active learning info. Defaults to True.
        Returns:
            list[DatasetItem]: List of dataset items
        """

        # Update info in case of change
        self.info = self.load_info()

        # Load PyArrow items from tables
        pyarrow_items: dict[str, dict[str, pa.Table]] = defaultdict(dict)

        # Search items with selected method
        if query["model"] in ["CLIP"]:
            pyarrow_items = self._embeddings_search(limit, offset, query)
        # NOTE: metadata search could go here

        if pyarrow_items is not None and pyarrow_items["main"]["db"].num_rows > 0:
            # Split results
            pyarrow_item_list = self._split_items(pyarrow_items, load_active_learning)

            return [
                DatasetItem.from_pyarrow(pyarrow_item, self.info, self.media_dir)
                for pyarrow_item in pyarrow_item_list
            ]
        return None

    def _embeddings_search(
        self,
        limit: int,
        offset: int,
        query: dict[str, str],
    ) -> dict[str, dict[str, pa.Table]]:
        """Perform item semantic search with embeddings

        Args:
            limit (int): Items limit
            offset (int): Items offset
            query (dict[str, str]): Search query

        Raises:
            ImportError: Required pixano-inference module could not be imported

        Returns:
            dict[str, dict[str, pa.Table]]: Search results
        """

        # Load tables
        ds_tables = self.open_tables()

        # Create PyArrow items
        pyarrow_items: dict[str, dict[str, pa.Table]] = defaultdict(dict)

        # Find CLIP embeddings
        if "embeddings" not in self.info.tables:
            return None
        for table in self.info.tables["embeddings"]:
            if table.type == "search" and table.source == query["model"]:
                sem_search_table = ds_tables["embeddings"][table.name]
                sem_search_views = [
                    field_name
                    for field_name, field_type in table.fields.items()
                    if field_type == "vector(512)"
                ]

        if query["model"] == "CLIP":
            # Initialize CLIP model
            try:
                # pylint: disable=import-outside-toplevel
                from pixano_inference.transformers import CLIP
            except ImportError as e:
                raise ImportError(
                    "Please install the pixano-inference module to perform semantic search with CLIP"
                ) from e

            model = CLIP()

        model_query = model.semantic_search(query["search"])

        # Perform semantic search
        # pylint: disable=unused-variable
        results_table = (
            sem_search_table.search(model_query, sem_search_views[0])
            .limit(min(offset + limit, self.num_rows))
            .to_arrow()
        )

        # If more than one view, search on all views and select the best results based on distance
        if len(sem_search_views) > 1:
            for view in sem_search_views[1:]:
                view_results_table = (
                    sem_search_table.search(model_query, view)
                    .limit(min(offset + limit, self.num_rows))
                    .to_arrow()
                )
                results_table = duckdb.query(
                    "SELECT id, results_table._distance as distance_1, view_results_table._distance as distance_2 FROM results_table LEFT JOIN view_results_table USING (id)"
                ).to_arrow_table()

                results_table = duckdb.query(
                    "SELECT (id), (SELECT Min(v) FROM (VALUES (distance_1), (distance_2)) AS value(v)) as _distance FROM results_table"
                ).to_arrow_table()

        # Filter results to page
        results_table = duckdb.query(
            f"SELECT id, _distance as distance FROM results_table ORDER BY _distance ASC LIMIT {limit} OFFSET {offset}"
        ).to_arrow_table()

        # Join with main table
        main_table = ds_tables["main"]["db"].to_lance()
        pyarrow_items["main"]["db"] = duckdb.query(
            "SELECT * FROM results_table LEFT JOIN main_table USING (id) ORDER BY distance ASC"
        ).to_arrow_table()

        return pyarrow_items

    def _split_items(
        self,
        pyarrow_items: dict[str, dict[str, pa.Table]],
        load_active_learning: bool,
    ) -> list[dict[str, dict[str, pa.Table]]]:
        """Split PyArrow tables into list of PyArrow tables

        Args:
            pyarrow_items (dict[str, dict[str, pa.Table]]): PyArrow tables
            load_active_learning (bool): Load items active learning info

        Returns:
            list[dict[str, dict[str, pa.Table]]]: List of PyArrow tables
        """

        # Load tables
        ds_tables = self.open_tables()

        # Create list of PyArrow tables
        pyarrow_item_list: list[dict[str, dict[str, pa.Table]]] = []

        for index in range(pyarrow_items["main"]["db"].num_rows):
            pyarrow_item_list.append(defaultdict(dict))
            # Main table
            pyarrow_item_list[index]["main"]["db"] = pyarrow_items["main"]["db"].take(
                [index]
            )
            item_id = pyarrow_item_list[index]["main"]["db"].to_pylist()[0]["id"]

            # Media tables
            for media_source, media_table in ds_tables["media"].items():
                # If media table already created
                if "media" in pyarrow_items:
                    pyarrow_item_list[index]["media"][media_source] = (
                        pa_ds.dataset(pyarrow_items["media"][media_source])
                        .scanner(filter=pa_ds.field("id") == item_id)
                        .to_table()
                    )
                # Else, retrieve media items individually
                else:
                    lance_scanner = media_table.to_lance().scanner(
                        filter=f"id in ('{item_id}')"
                    )
                    pyarrow_item_list[index]["media"][
                        media_source
                    ] = lance_scanner.to_table()

            # Active learning tables
            if load_active_learning:
                # If active learning table already created
                if "active_learning" in pyarrow_items:
                    for al_source in ds_tables["active_learning"].keys():
                        pyarrow_item_list[index]["active_learning"][al_source] = (
                            pa_ds.dataset(pyarrow_items["active_learning"][al_source])
                            .scanner(filter=pa_ds.field("id") == item_id)
                            .to_table()
                        )
                # Else, retrieve active learning items individually
                else:
                    for al_source, al_table in ds_tables["active_learning"].items():
                        lance_scanner = al_table.to_lance().scanner(
                            filter=f"id in ('{item_id}')"
                        )
                        pyarrow_item_list[index]["active_learning"][
                            al_source
                        ] = lance_scanner.to_table()

        return pyarrow_item_list

    def load_item(
        self,
        item_id: str,
        load_media: bool = True,
        load_objects: bool = False,
        load_active_learning: bool = True,
        load_embeddings: bool = False,
        model_id: str = None,
    ) -> DatasetItem:
        """Find dataset item in selected tables

        Args:
            item_id (str): Dataset item ID
            load_media (bool, optional): Load item media. Defaults to True.
            load_objects (bool, optional): Load item objects. Defaults to False.
            load_active_learning (bool, optional): Load item active learning info. Defaults to True.
            load_embeddings (bool, optional): Load item embeddings. Defaults to False.
            model_id (str, optional): Model ID (ONNX file path) of embeddings to load. Defaults to None.
        Returns:
            DatasetItem: Dataset item
        """

        # Update info in case of change
        self.info = self.load_info()

        # Load tables
        ds_tables = self.open_tables()

        # Load PyArrow item from tables
        pyarrow_item: dict[str, dict[str, pa.Table]] = defaultdict(dict)

        # Load PyArrow item from main table
        lance_scanner = (
            ds_tables["main"]["db"].to_lance().scanner(filter=f"id in ('{item_id}')")
        )
        pyarrow_item["main"]["db"] = lance_scanner.to_table()

        # Load PyArrow item from media tables
        if load_media:
            for table_name, media_table in ds_tables["media"].items():
                lance_scanner = media_table.to_lance().scanner(
                    filter=f"id in ('{item_id}')"
                )
                pyarrow_item["media"][table_name] = lance_scanner.to_table()

        # Load PyArrow item from objects tables
        if load_objects:
            for table_name, obj_table in ds_tables["objects"].items():
                lance_scanner = obj_table.to_lance().scanner(
                    filter=f"item_id in ('{item_id}')"
                )
                pyarrow_item["objects"][table_name] = lance_scanner.to_table()

        # Load PyArrow item from active learning tables
        if load_active_learning:
            for table_name, al_table in ds_tables["active_learning"].items():
                lance_scanner = al_table.to_lance().scanner(
                    filter=f"id in ('{item_id}')"
                )
                pyarrow_item["active_learning"][table_name] = lance_scanner.to_table()

        # Load PyArrow item from segmentation embeddings tables
        found_embeddings = not load_embeddings
        if load_embeddings and "embeddings" in self.info.tables:
            for table in self.info.tables["embeddings"]:
                if table.source.lower() in model_id.lower():
                    found_embeddings = True
                    emb_table = ds_tables["embeddings"][table.name]
                    lance_scanner = emb_table.to_lance().scanner(
                        filter=f"id in ('{item_id}')"
                    )
                    pyarrow_item["embeddings"][table.name] = lance_scanner.to_table()

        if pyarrow_item["main"]["db"].num_rows > 0 and found_embeddings:
            return DatasetItem.from_pyarrow(
                pyarrow_item,
                self.info,
                self.media_dir,
                media_features=True,
                model_id=model_id,
            )
        return None

    def save_item(self, item: DatasetItem):
        """Save dataset item features and objects

        Args:
            item (DatasetItem): Item to save
        """

        # Update info in case of change
        self.info = self.load_info()

        # Load dataset tables
        ds_tables = self.open_tables()
        main_table = ds_tables["main"]["db"]

        # Add new item columns
        self.update_table(item, main_table, "main", "db")

        # Reload dataset tables
        ds_tables = self.open_tables()
        main_table = ds_tables["main"]["db"]

        # Update item
        item.update(main_table)

        # Add or update item objects
        for obj in item.objects.values():
            table_found = False
            if "objects" in self.info.tables:
                for table in self.info.tables["objects"]:
                    if table.source == obj.source_id:
                        # Load object table
                        table_found = True
                        obj_table = ds_tables["objects"][table.name]

                        # Add new object columns
                        self.update_table(obj, obj_table, "objects", table.name)

                        # Reload dataset tables
                        ds_tables = self.open_tables()
                        main_table = ds_tables["objects"][table.name]

                        # Add or update object
                        obj.add_or_update(obj_table)
            # If first object
            if not table_found and obj.source_id == "Ground Truth":
                # Create table
                table = DatasetTable(
                    name="objects",
                    source="Ground Truth",
                    fields={
                        "id": "str",
                        "item_id": "str",
                        "view_id": "str",
                        "bbox": "bbox",
                        "mask": "compressedrle",
                    },
                )
                for feat in obj.features.values():
                    table.fields[feat.name] = feat.dtype
                self.create_table(table, "objects")

                # Reload dataset tables
                ds_tables = self.open_tables()
                obj_table = ds_tables["objects"][table.name]

                # Add object
                obj.add_or_update(obj_table)

        # Delete removed item objects
        item.delete_objects(ds_tables)

    def get_features_values(self, config_values: FeaturesValues) -> FeaturesValues:
        """get config values
           merge with distinct existing values for each scene and object string features, if not restricted

        Args:
            config_values (FeaturesValues): features values from db.json
        Returns:
            FeaturesValues: existing values for each scene and object string feature
        """

        # Load tables
        ds_tables = self.open_tables()

        def get_distinct_values(
            table_name: str,
            ignore_list: list[str],
            config_vals: dict[str, FeatureValues],
        ) -> dict[str, FeatureValues]:
            avail_values = defaultdict(FeatureValues)
            for table in ds_tables[table_name].values():
                table_arrow = table.to_arrow()
                feats = [f for f in table_arrow.column_names if f not in ignore_list]
                for feat in feats:
                    if feat not in avail_values:
                        avail_values[feat] = FeatureValues(restricted=False, values=[])
                    if config_vals and feat in config_vals:
                        avail_values[feat].restricted = config_vals[feat].restricted
                        avail_values[feat].values.extend(config_vals[feat].values)
                    if not avail_values[feat].restricted:
                        v = (
                            duckdb.sql(f"select DISTINCT {feat} from table_arrow")
                            .to_arrow_table()
                            .to_pydict()
                        )
                        if v[feat] is not None and v[feat] != [None]:
                            avail_values[feat].values.extend(
                                [
                                    val
                                    for val in v[feat]
                                    if val is not None
                                    and isinstance(val, str)
                                    and val not in avail_values[feat].values
                                ]
                            )
            return avail_values

        return FeaturesValues(
            main=get_distinct_values(
                "main",
                ["id", "split", "views", "original_id"],
                config_values.main if config_values else None,
            ),
            objects=get_distinct_values(
                "objects",
                ["id", "item_id", "view_id", "bbox", "mask", "review_state"],
                config_values.objects if config_values else None,
            ),
        )

    @staticmethod
    def find(
        dataset_id: str,
        directory: Path | S3Path,
    ) -> "Dataset":
        """Find Dataset in directory

        Args:
            dataset_id (str): Dataset ID
            directory (Path): Directory to search in

        Returns:
            Dataset: Dataset
        """

        # Browse directory
        for json_fp in directory.glob("*/db.json"):
            info = DatasetInfo.from_json(json_fp)
            if info.id == dataset_id:
                # Return dataset
                return Dataset(json_fp.parent)
        return None



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Optional

import pyarrow as pa
from pydantic import BaseModel

from pixano.core.pixano_type import PixanoType, create_pyarrow_type


class Camera(PixanoType, BaseModel):
    """Camera type

    Attributes:
        depth_scale (float): Depth scale
        cam_k (list[float]): Camera matrix K
        cam_r_w2c (list[float], optional): 3*3 orientation matrix
        cam_t_w2c (list[float], optional): 3*1 translation matrix
    """

    depth_scale: float
    cam_k: list[float]
    cam_r_w2c: Optional[list[float]]
    cam_t_w2c: Optional[list[float]]

    def __init__(
        self,
        depth_scale: float,
        cam_k: list[float],
        cam_r_w2c: list[float] = None,
        cam_t_w2c: list[float] = None,
    ):
        """Initialize Camera

        Args:
            depth_scale (float): Depth scale
            cam_k (list[float]): Camera matrix K
            cam_r_w2c (list[float], optional): 3*3 orientation matrix. Defaults to None.
            cam_t_w2c (list[float], optional): 3*1 translation matrix. Defaults to None.
        """

        if cam_r_w2c is None:
            cam_r_w2c = [0.0] * 9
        if cam_t_w2c is None:
            cam_t_w2c = [0.0] * 3

        # Define public attributes through Pydantic BaseModel
        super().__init__(
            depth_scale=depth_scale,
            cam_k=cam_k,
            cam_r_w2c=cam_r_w2c,
            cam_t_w2c=cam_t_w2c,
        )

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return Camera type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("depth_scale", pa.float64()),
                pa.field("cam_k", pa.list_(pa.float64())),
                pa.field("cam_r_w2c", pa.list_(pa.float64())),
                pa.field("cam_t_w2c", pa.list_(pa.float64())),
            ]
        )


CameraType = create_pyarrow_type(Camera.to_struct(), "Camera", Camera)



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from abc import ABC, abstractmethod
from typing import Any, Type

import pyarrow as pa
from pydantic import BaseModel


class PixanoType(ABC, BaseModel):
    """Base class for all Pixano custom types"""

    def to_dict(self) -> dict[str, Any]:
        """Return custom type as dict based on corresponding PyArrow Struct

        Returns:
            dict[str, Any]: Custom type as dict
        """

        def _convert_value_as_dict(value):
            """Recursively convert value to dict if possible"""

            if isinstance(value, PixanoType):
                return value.to_dict()
            if isinstance(value, dict):
                return {k: _convert_value_as_dict(v) for k, v in value.items()}
            if isinstance(value, (list, tuple)):
                return [_convert_value_as_dict(item) for item in value]
            return value

        struct_fields = self.to_struct()
        return {
            field.name: _convert_value_as_dict(getattr(self, field.name))
            for field in struct_fields
        }

    @classmethod
    def from_dict(cls: Type["PixanoType"], data: dict[str, Any]) -> "PixanoType":
        """Instance custom type from dict

        Args:
            cls (Type[PixanoType]): Pixano custom type to instance
            data (dict[str, Any]): Data to instance from

        Returns:
            PixanoType: New instance of Pixano custom type
        """

        return cls(**data)

    @staticmethod
    @abstractmethod
    def to_struct() -> pa.StructType:
        """Return custom type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """


def convert_field(
    field_name: str, field_type: pa.DataType, field_data: list
) -> pa.Array:
    """Convert PyArrow ExtensionTypes properly

    Args:
        field_name (str): Name
        field_type (pa.DataType): Target PyArrow format
        field_data (list): Data in Python format

    Returns:
        pa.Array: Data in target PyArrow format
    """

    # If target format is an ExtensionType
    if isinstance(field_type, pa.ExtensionType):
        storage = pa.array(field_data, type=field_type.storage_type)
        return pa.ExtensionArray.from_storage(field_type, storage)

    # If target format is an extension of ListType
    if pa.types.is_list(field_type):
        native_arr = pa.array(field_data)
        if isinstance(native_arr, pa.NullArray):
            return pa.nulls(len(native_arr), field_type)
        offsets = native_arr.offsets
        values = native_arr.values.to_numpy(zero_copy_only=False)
        return pa.ListArray.from_arrays(
            offsets,
            convert_field(f"{field_name}.elements", field_type.value_type, values),
        )

    # If target format is an extension of StructType
    if pa.types.is_struct(field_type):
        native_arr = pa.array(field_data)
        if isinstance(native_arr, pa.NullArray):
            return pa.nulls(len(native_arr), field_type)
        arrays = []
        for subfield in field_type:
            sub_arr = native_arr.field(subfield.name)
            converted = convert_field(
                f"{field_name}.{subfield.name}",
                subfield.type,
                sub_arr.to_numpy(zero_copy_only=False),
            )
            arrays.append(converted)
        return pa.StructArray.from_arrays(arrays, fields=field_type)

    # For other target formats
    return pa.array(field_data, type=field_type)


def create_pyarrow_type(
    struct_type: pa.StructType,
    name: str,
    python_type: type,
) -> pa.ExtensionType:
    """Create PyArrow ExtensionType for Pixano custom type

    Args:
        struct_type (pa.StructType): Pixano custom type as PyArrow Struct
        name (str): Pixano custom type name
        python_type (type): Pixano custom type Python type

    Returns:
        pa.ExtensionType: PyArrow ExtensionType
    """

    class CustomExtensionType(pa.ExtensionType):
        """PyArrow CustomExtensionType"""

        # pylint: disable=unused-argument
        @classmethod
        def __arrow_ext_deserialize__(cls, storage_type, serialized):
            return cls(struct_type, name)

        def __arrow_ext_serialize__(self):
            return b""

        def __arrow_ext_scalar_class__(self):
            return self.Scalar

        def __arrow_ext_class__(self):
            return self.Array

        def __repr__(self):
            return f"ExtensionType<{name}Type>"

        class Scalar(pa.ExtensionScalar):
            """PyArrow ExtensionScalar"""

            def as_py(self):
                """Create Python PixanoType from ExtensionScalar dictionary of PyArrow objects"""

                def as_py_dict(pa_dict: dict[str, Any]) -> dict[str, Any]:
                    """Recusively convert dictionary of PyArrow objects to dictionary of Python objects

                    Args:
                        pa_dict (dict[str, Any]): Dictionary of PyArrow objects

                    Returns:
                        dict[str, Any]: Dictionary of Python objects
                    """

                    py_dict = {}
                    for key, value in pa_dict.items():
                        if hasattr(value, "as_py") and callable(
                            getattr(value, "as_py")
                        ):
                            py_dict[key] = value.as_py()
                        elif isinstance(value, dict):
                            py_dict[key] = as_py_dict(value)
                    return py_dict

                return python_type.from_dict(as_py_dict(self.value))

        class Array(pa.ExtensionArray):
            """PyArrow ExtensionArray"""

            def __repr__(self):
                return f"<{name}Array object at {hex(id(self))}>\n{self}"

            @staticmethod
            def from_pylist(lst: list) -> pa.ExtensionArray | pa.ListArray:
                """Create PyArrow list of objects from Python list


                Args:
                    lst (list): Python list

                Returns:
                    pa.ExtensionArray | pa.ListArray: PyArrow list
                """

                def from_list(lst: list) -> pa.ExtensionArray:
                    """Return ExtensionArray from Python list

                    Args:
                        lst (list): Python list

                    Returns:
                        pa.ExtensionArray: ExtensionArray
                    """

                    fields = struct_type
                    arrays = []

                    for field in fields:
                        data = []
                        for obj in lst:
                            if obj is not None:
                                if hasattr(obj, "to_dict") and callable(
                                    getattr(obj, "to_dict")
                                ):
                                    data.append(obj.to_dict().get(field.name))
                                elif isinstance(obj, dict):
                                    data.append(obj.get(field.name))
                                else:
                                    data.append(obj)
                            else:
                                data.append(None)

                        arrays.append(
                            convert_field(
                                field.name,
                                field.type,
                                data,
                            )
                        )
                    sto = pa.StructArray.from_arrays(arrays, fields=fields)
                    return pa.ExtensionArray.from_storage(pyarrow_type, sto)

                def from_lists(lsts: list[list]) -> pa.ListArray:
                    """Return ListArray from Python list of lists

                    Args:
                        lsts (list[list]): Python list of lists

                    Returns:
                        pa.ListArray: ListArray
                    """

                    offset = [0]
                    for sub_list in lsts:
                        offset.append(len(sub_list) + offset[-1])

                    flat_list = [item for sublist in lsts for item in sublist]
                    flat_array = from_list(flat_list)

                    return pa.ListArray.from_arrays(
                        offset, flat_array, type=pa.list_(pyarrow_type)
                    )

                def is_nested(lst: list) -> bool:
                    """Check if list contains only sublists

                    Args:
                        lst (list): List to check

                    Returns:
                        bool: True if list contains only sublists
                    """

                    return all(isinstance(item, list) for item in lst)

                def is_flat(lst: list) -> bool:
                    """Check if list does not contain sublists

                    Args:
                        lst (list): List to check

                    Returns:
                        bool: True if list does not contains sublists
                    """

                    return all(not isinstance(item, list) for item in lst)

                if is_nested(lst):
                    return from_lists(lst)
                if is_flat(lst):
                    return from_list(lst)
                raise ValueError(
                    "Input list must be either a nested list or a flat list"
                )

    # Create ExtensionType
    pyarrow_type = CustomExtensionType(struct_type, name)

    # Try and register ExtensionType
    try:
        pa.register_extension_type(pyarrow_type)
    # If ExtensionType is already registered
    except pa.ArrowKeyError:
        pass

    return pyarrow_type



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from pathlib import Path
from typing import IO, Optional
from urllib.parse import urlparse
from urllib.request import urlopen

import cv2
import numpy as np
import pyarrow as pa
from IPython.core.display import Image as IPyImage
from PIL import Image as PILImage
from pydantic import BaseModel
from s3path import S3Path

from pixano.core.pixano_type import PixanoType, create_pyarrow_type
from pixano.utils import binary_to_url

# Disable warning for Image "bytes" attribute
# NOTE: Rename attribute? Breaking change for Pixano datasets using Image
# pylint: disable=redefined-builtin, used-before-assignment


class Image(PixanoType, BaseModel):
    """Image type using URI or bytes

    Attributes:
        uri (str): Image URI
        bytes (bytes): Image bytes
        preview_bytes (bytes): Image preview bytes
        uri_prefix (str): URI prefix for relative URIs
    """

    uri: str
    bytes: Optional[bytes]
    preview_bytes: Optional[bytes]
    uri_prefix: Optional[str]

    def __init__(
        self,
        uri: str,
        bytes: bytes = None,
        preview_bytes: bytes = None,
        uri_prefix: str = None,
    ):
        """Initialize Image

        Attributes:
            uri (str): Image URI
            bytes (bytes, optional): Image bytes. Defaults to None.
            preview_bytes (bytes, optional): Image preview bytes. Defaults to None.
            uri_prefix (str, optional): URI prefix for relative URIs. Defaults to None.
        """

        # Define public attributes through Pydantic BaseModel
        super().__init__(
            uri=uri,
            bytes=bytes,
            preview_bytes=preview_bytes,
            uri_prefix=uri_prefix,
        )

    @property
    def complete_uri(self) -> str:
        """Return complete image URI using URI and URI prefix

        Returns:
            str: Image URI
        """

        # URI is incomplete
        if urlparse(self.uri).scheme == "":
            # URI prefix exists
            if self.uri_prefix is not None:
                parsed_uri = urlparse(self.uri_prefix)
                # URI prefix is incomplete
                if parsed_uri.scheme == "":
                    raise ValueError(
                        "URI prefix is incomplete, no scheme provided (http://, file://, ...)"
                    )
                combined_path = Path(parsed_uri.path) / self.uri
                parsed_uri = parsed_uri._replace(path=str(combined_path))
                return parsed_uri.geturl()
            # No URI prefix
            raise ValueError("URI is incomplete, need URI prefix")
        # URI is already complete
        return self.uri

    @property
    def path(self) -> Path:
        """Return image path using URI and URI prefix

        Returns:
            Path: Image path
        """

        path_from_uri = urlparse(self.complete_uri).path

        return Path(path_from_uri)

    @property
    def url(self) -> str:
        """Return image base 64 URL

        Returns:
            str: Image base 64 URL
        """

        return binary_to_url(self.get_bytes())

    @property
    def preview_url(self) -> str:
        """Return image preview base 64 URL

        Returns:
            str: Image preview base 64 URL
        """

        return binary_to_url(self.preview_bytes)

    @property
    def file_name(self) -> str:
        """Return image file name from URI

        Returns:
            str: Image file name
        """

        return Path(urlparse(self.uri).path).name

    @property
    def size(self) -> list[int]:
        """Return image size

        Returns:
            list[int]: Image size
        """

        return self.as_pillow().size

    @property
    def width(self) -> int:
        """Return image width

        Returns:
            int: Image width
        """

        return self.as_pillow().width

    @property
    def height(self) -> int:
        """Return image height

        Returns:
            int: Image height
        """

        return self.as_pillow().height

    def get_bytes(self) -> bytes:
        """Get image bytes from attribute or from reading file from URI

        Returns:
            bytes: Image bytes
        """

        if self.bytes is not None:
            return self.bytes
        if self.uri is not None:
            with self.open() as f:
                return f.read()
        return None

    def open(self) -> IO:
        """Open image

        Returns:
            IO: Opened image
        """

        complete_uri = self.complete_uri
        if urlparse(complete_uri).scheme == "s3":
            presigned_url = S3Path.from_uri(complete_uri).get_presigned_url()
            return urlopen(presigned_url)
        return urlopen(complete_uri)

    def as_pillow(self) -> PILImage.Image:
        """Open image as Pillow

        Returns:
            PIL.Image.Image: Image as Pillow
        """

        return PILImage.open(self.open()).convert("RGB")

    def as_cv2(self) -> np.ndarray:
        """Open image as OpenCV

        Returns:
            np.ndarray: Image as OpenCV
        """

        im_arr = np.frombuffer(self.open().read(), dtype=np.uint8)
        return cv2.imdecode(im_arr, cv2.IMREAD_COLOR)

    def display(self, preview=False) -> IPyImage:
        """Display image

        Args:
            preview (bool, optional): True to display image preview instead of full image. Defaults to False.

        Returns:
            IPython.core.display.Image: Image as IPython Display
        """

        im_bytes = self.preview_bytes if preview else self.get_bytes()
        return IPyImage(url=binary_to_url(im_bytes), format=IPyImage(im_bytes).format)

    @staticmethod
    def to_struct() -> pa.StructType:
        """Return Image type as PyArrow Struct

        Returns:
            pa.StructType: Custom type corresponding PyArrow Struct
        """

        return pa.struct(
            [
                pa.field("uri", pa.utf8()),
                pa.field("bytes", pa.binary()),
                pa.field("preview_bytes", pa.binary()),
            ]
        )


ImageType = create_pyarrow_type(Image.to_struct(), "Image", Image)



# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import asyncio
from functools import lru_cache

import click
import fastapi
import pkg_resources
import uvicorn
from fastapi.responses import HTMLResponse
from fastapi.staticfiles import StaticFiles
from fastapi.templating import Jinja2Templates

from pixano.app.display import display_cli, display_colab, display_ipython
from pixano.app.main import create_app
from pixano.data.settings import Settings, get_settings

LOGO = """
                             ▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▓▒
                           ▒▓▒                                       ▓
              █▓          ▓▓                                         ▓
                        ▒▓▒   ▒▓▓▓▓▓ ▒▓   ▓▒▒▓▓▓▓▒     ▒▓▓▓▓▓▒       ▓
▓▓ ▓▓▓▓▓▓▒    ▓▒  ▒▓   ▓▒   ▒▓      ▓██   ██     ▓▒  ▒█▓     ▒█▓     ▓
███      ▓▓   █▒   ▓█      ▒█▒       ██   █▒     ▒█  ▓▒        █     ▓
██        █▒  █▒    ▒█▓     █▒       ██   █▒     ▒█  ▓▓        █     ▓
██        █▒  █▒    ▓█▓█     █▒     ▓██   █▒     ▒█   ██     ▓█▒     ▓
██▓▒    ▒▓▒   █▒   ▓▒  ▒█▓    ▒▓▓▓▓▓ ▒▓   ▓      ▒▓    ▒▓▓▓▓▓▒       ▓
██  ▒▓▓▒▒     ▒   ▒      ▒▓▒                                         ▓
██                         █▓                                        ▓
"""
ASSETS_PATH = pkg_resources.resource_filename("pixano", "app/dist/_app")
TEMPLATE_PATH = pkg_resources.resource_filename("pixano", "app/dist")

task_functions = {
    "colab": asyncio.get_event_loop().create_task,
    "ipython": asyncio.get_event_loop().create_task,
    "none": asyncio.run,
}
display_functions = {
    "colab": display_colab,
    "ipython": display_ipython,
    "none": display_cli,
}


class App:
    """Pixano app

    Attributes:
        app (fastapi.FastAPI): FastAPI App
        config (uvicorn.Config): App config
        server (uvicorn.Server): App server
    """

    app: fastapi.FastAPI
    config: uvicorn.Config
    server: uvicorn.Server

    def __init__(
        self,
        library_dir: str,
        aws_endpoint: str = None,
        aws_region: str = None,
        aws_access_key: str = None,
        aws_secret_key: str = None,
        local_model_dir: str = None,
        host: str = "127.0.0.1",
        port: int = 8000,
    ):
        """Initialize and run Pixano app

        Args:
            library_dir (str): Local or S3 path to dataset library
            aws_endpoint (str, optional): S3 endpoint URL, use 'AWS' if not provided. Used if library_dir is an S3 path. Defaults to None.
            aws_region (str, optional): S3 region name, not always required for private storages. Used if library_dir is an S3 path. Defaults to None.
            aws_access_key (str, optional): S3 AWS access key. Used if library_dir is an S3 path. Defaults to None.
            aws_secret_key (str, optional): S3 AWS secret key. Used if library_dir is an S3 path. Defaults to None.
            local_model_dir (str, optional): Local path to models. Used if library_dir is an S3 path. Defaults to None.
            host (str, optional): App host. Defaults to "127.0.0.1".
            port (int, optional): App port. Defaults to 8000.
        """

        # Override app settings
        @lru_cache
        def get_settings_override():
            return Settings(
                library_dir=library_dir,
                aws_endpoint=aws_endpoint,
                aws_region=aws_region,
                aws_access_key=aws_access_key,
                aws_secret_key=aws_secret_key,
                local_model_dir=local_model_dir,
            )

        # Create app
        templates = Jinja2Templates(directory=TEMPLATE_PATH)
        self.app = create_app(settings=get_settings_override())
        self.app.dependency_overrides[get_settings] = get_settings_override

        @self.app.get("/", response_class=HTMLResponse)
        def main_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        @self.app.get("/{ds_id}/dataset", response_class=HTMLResponse)
        async def dataset_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        @self.app.get("/{ds_id}/dashboard", response_class=HTMLResponse)
        async def dashboard_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        @self.app.get("/{ds_id}/dataset/{item_id}", response_class=HTMLResponse)
        async def item_page(request: fastapi.Request):
            return templates.TemplateResponse("index.html", {"request": request})

        self.app.mount("/_app", StaticFiles(directory=ASSETS_PATH), name="assets")
        self.config = uvicorn.Config(self.app, host=host, port=port)
        self.server = uvicorn.Server(self.config)

        # Serve app
        task_functions[self.get_env()](self.server.serve())

    def display(self, height: int = 1000) -> None:
        """Display Pixano app

        Args:
            height (int, optional): Frame height. Defaults to 1000.
        """

        # Wait for app to be online
        while not self.server.started:
            task_functions[self.get_env()](asyncio.wait(0.1))

        # Display app
        for server in self.server.servers:
            for socket in server.sockets:
                address = socket.getsockname()
                display_functions[self.get_env()](
                    url=f"http://{address[0]}", port=address[1], height=height
                )

    def get_env(self) -> str:
        """Get the app's current running environment

        Returns:
            str: Running environment
        """

        # If Google colab import succeeds
        try:
            # pylint: disable=import-outside-toplevel, unused-import
            import google.colab
            import IPython
        except ImportError:
            pass
        else:
            if IPython.get_ipython() is not None:
                return "colab"

        # If IPython import succeeds
        try:
            # pylint: disable=import-outside-toplevel
            import IPython
        except ImportError:
            pass
        else:
            ipython = IPython.get_ipython()
            if ipython is not None and ipython.has_trait("kernel"):
                return "ipython"

        # Else
        return "none"


@click.command(context_settings={"auto_envvar_prefix": "UVICORN"})
@click.argument("library_dir", type=str)
@click.option(
    "--aws_endpoint",
    type=str,
    help="S3 endpoint URL, use 'AWS' if not provided. Used if library_dir is an S3 path",
)
@click.option(
    "--aws_region",
    type=str,
    help="S3 region name, not always required for private storages. Used if library_dir is an S3 path",
)
@click.option(
    "--aws_access_key",
    type=str,
    help="S3 AWS access key. Used if library_dir is an S3 path",
)
@click.option(
    "--aws_secret_key",
    type=str,
    help="S3 AWS secret key. Used if library_dir is an S3 path",
)
@click.option(
    "--local_model_dir",
    type=str,
    help="Local path to your models. Used if library_dir is an S3 path",
)
@click.option(
    "--host",
    type=str,
    default="127.0.0.1",
    help="Pixano app URL host",
    show_default=True,
)
@click.option(
    "--port",
    type=int,
    default=0,
    help="Pixano app URL port",
    show_default=True,
)
def main(
    library_dir: str,
    aws_endpoint: str,
    aws_region: str,
    aws_access_key: str,
    aws_secret_key: str,
    local_model_dir: str,
    host: str,
    port: int,
):
    """Launch Pixano App in LIBRARY_DIR

    LIBRARY_DIR is the local or S3 path to your dataset library
    """

    App(
        library_dir,
        aws_endpoint,
        aws_region,
        aws_access_key,
        aws_secret_key,
        local_model_dir,
        host,
        port,
    )



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Optional

import pyarrow as pa
from pydantic import BaseModel

from pixano.core import is_boolean, is_float, is_integer, is_string


class ItemFeature(BaseModel):
    """Feature

    Attributes:
        name (str): Feature name
        dtype (str): Feature type
        value (str | int | float | bool, optional): Feature value
    """

    name: str
    dtype: str
    value: Optional[str | int | float | bool] = None

    @staticmethod
    def from_pyarrow(
        table: pa.Table,
        schema: pa.schema,
    ) -> dict[str, "ItemFeature"]:
        """Create dictionary of ItemFeature from PyArrow Table

        Args:
            table (pa.Table): PyArrow table
            schema (pa.schema): PyArrow schema

        Returns:
            dict[str, ItemFeature]: Dictionary of ItemFeature
        """

        item = table.to_pylist()[0]
        features = {}
        ignored_fields = [
            "id",
            "item_id",
            "view_id",
            "source_id",
            "split",
            "review_state",
        ]

        # Iterate on fields
        for field in schema:
            if field.name not in ignored_fields:
                # Integer fields
                if is_integer(field.type):
                    features[field.name] = ItemFeature(
                        name=field.name,
                        dtype="int",
                        value=item[field.name],
                    )

                # Float fields
                if is_float(field.type):
                    # Parse float value from string
                    # (Float conversions from PyArrow to Python can currently add a lot of random decimal places)
                    value_as_string: str = table[field.name].to_string()
                    value_as_string = (
                        value_as_string.replace("[", "").replace("]", "").strip()
                    )
                    try:
                        features[field.name] = ItemFeature(
                            name=field.name,
                            dtype="float",
                            value=float(value_as_string),
                        )
                    except ValueError:
                        features[field.name] = ItemFeature(
                            name=field.name,
                            dtype="float",
                            value=float(item[field.name]),
                        )

                # String fields
                elif is_string(field.type):
                    features[field.name] = ItemFeature(
                        name=field.name,
                        dtype="str",
                        value=str(item[field.name]),
                    )

                # Boolean fields
                elif is_boolean(field.type):
                    features[field.name] = ItemFeature(
                        name=field.name,
                        dtype="bool",
                        value=bool(item[field.name]),
                    )

        # Additional distance field in case of semantic search
        for field_name in item.keys():
            if field_name == "distance":
                features["search distance"] = ItemFeature(
                    name="search distance",
                    dtype="float",
                    value=round(item[field_name], 2),
                )

        return features


class FeatureValues(BaseModel):
    """Feature available values, as a restricted list or open list

    Attributes:
        restricted (bool): restricted list or open list
        values (list[str | int]): list of available values
    """

    restricted: bool
    values: list[str | int]


class FeaturesValues(BaseModel):
    """Features availables values

    Attributes:
        main (dict[str, FeatureValues], optional): Scene features available values ("main" table)
        objects (dict[str, FeatureValues], optional): Objects features available values
    """

    main: Optional[dict[str, FeatureValues]] = None
    objects: Optional[dict[str, FeatureValues]] = None



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from fractions import Fraction

import pyarrow as pa
from PIL import Image as PILImage

from pixano.data import Dataset


def compute_image_stats(ds: Dataset):
    """Compute image stats, save them to stats.json

    Args:
        ds (Dataset): Dataset
    """

    tables = ds.open_tables()

    for view in tables["media"]:
        # will be flattened, so don't treat it as a real loop (only one elem)
        # tt = tables["media"][view].to_lance()
        # print(duckdb.sql("select * from tt"))
        data_table = tables["media"][view].to_arrow()

        # Take a subset of table without image columns (which can't be converted to pandas)
        if not all(p in data_table.column_names for p in ["width", "height"]):
            print(
                "INFO: 'width' and 'height' not found in media table, get it from image"
            )
            images = data_table.select([view]).to_pylist()
            sizes = []
            for image in images:
                # im = image[view].as_pillow() ne marche plus car uri_prefix vide (pb avec Image.get_uri())
                im = PILImage.open(ds.media_dir / image[view].uri)
                sizes.append({"width": im.width, "height": im.height})
            data = pa.Table.from_pylist(sizes).to_pandas()
        else:
            print("INFO: 'width' and 'height' found in media table, use it")
            data = data_table.select(["width", "height"]).to_pandas()

        # Compute additional data
        data["resolution"] = data.apply(
            lambda x: str(x["width"]) + "x" + str(x["height"]), axis=1
        )
        data["aspect_ratio"] = data.apply(
            lambda x: str(Fraction(x["width"], x["height"])).replace("/", ":"), axis=1
        )
        return data



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import pyarrow as pa

from pixano.core.image import ImageType


def is_integer(t: pa.DataType) -> bool:
    """Check if DataType is an integer

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is an integer
    """

    return pa.types.is_integer(t)


def is_float(t: pa.DataType) -> bool:
    """Check if DataType is a float

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is a float
    """

    return pa.types.is_floating(t)


def is_string(t: pa.DataType) -> bool:
    """Check if DataType is a string

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is a string
    """

    return pa.types.is_string(t) or pa.types.is_large_string(t)


def is_boolean(t: pa.DataType) -> bool:
    """Check if DataType is boolean

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is boolean
    """

    return pa.types.is_boolean(t)


def is_binary(t: pa.DataType) -> bool:
    """Check if DataType is binary

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is binary
    """

    return pa.types.is_binary(t)


def is_image_type(t: pa.DataType) -> bool:
    """Check if DataType is an Image

    Args:
        t (pa.DataType): DataType to check

    Returns:
        bool: True if DataType is an Image
    """

    return (
        ImageType.equals(t)
        or str(t) == "struct<uri: string, bytes: binary, preview_bytes: binary>"
    )



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

from typing import Any, Optional

import lancedb
import pyarrow as pa
from pydantic import BaseModel

from pixano.core import BBox, CompressedRLE
from pixano.data.fields import field_to_python
from pixano.data.item.item_feature import ItemFeature


class ItemURLE(BaseModel):
    """Uncompressed URLE mask type for DatasetItem

    Type for CompressedRLE.to_urle()

    Attributes:
        size (list[int]): Mask size
        counts (list[int]): Mask URLE encoding
    """

    size: list[int]
    counts: list[int]

    @staticmethod
    def from_pyarrow(
        rle: CompressedRLE,
    ) -> "ItemURLE":
        """Create ItemURLE from compressed RLE

        Args:
            rle (CompressedRLE): Compressed RLE

        Returns:
            ItemURLE: ItemURLE
        """

        return ItemURLE.model_validate(rle.to_urle()) if rle.counts else None

    def to_pyarrow(self) -> CompressedRLE:
        """Return ItemURLE as compressed RLE

        Returns:
            CompressedRLE: Compressed RLE
        """

        return CompressedRLE.from_urle(self.model_dump()) if self.counts else None


class ItemBBox(BaseModel):
    """BBox type for DatasetItem

    Type for BBox.to_dict()

    Attributes:
        coords (list[float]): List of coordinates in given format
        format (str): Coordinates format, 'xyxy' or 'xywh'
        is_normalized (bool, optional): True if coordinates are normalized to image size
        confidence (float, optional): Bounding box confidence if predicted
    """

    coords: list[float]
    format: str
    is_normalized: Optional[bool] = None
    confidence: Optional[float] = None

    @staticmethod
    def from_pyarrow(
        bbox: BBox,
    ) -> "ItemBBox":
        """Create ItemBBox from bounding box

        Args:
            bbox (BBox): Bounding box

        Returns:
            ItemBBox: ItemBBox
        """

        return (
            ItemBBox.model_validate(bbox.to_xywh().to_dict())
            if bbox.coords != [0.0, 0.0, 0.0, 0.0]
            else None
        )

    def to_pyarrow(self) -> BBox:
        """Return ItemBbox as BBox

        Returns:
            BBox: Bounding box
        """

        return (
            BBox.from_dict(self.model_dump())
            if self.coords != [0.0, 0.0, 0.0, 0.0]
            else None
        )


class ItemObject(BaseModel):
    """Object type for DatasetItem

    Attributes:
        id (str): Object ID
        original_id (str, optional): Object original ID
        item_id (str): Object item ID
        view_id (str): Object view ID
        source_id (str): Object source ID
        review_state (str, optional): Object review state ("accepted", "rejected", None)
        bbox (ItemBBox, optional): Object bounding box
        mask (ItemURLE, optional): Object mask
        features (dict[str, ItemFeature], optional): Object features
    """

    id: str
    original_id: Optional[str] = None
    item_id: str
    view_id: str
    source_id: str
    review_state: Optional[str] = None
    bbox: Optional[ItemBBox] = None
    mask: Optional[ItemURLE] = None
    features: Optional[dict[str, ItemFeature]] = None

    @staticmethod
    def from_pyarrow(
        table: pa.Table,
        schema: pa.schema,
        source_id: str,
    ) -> dict[str, "ItemObject"]:
        """Create dictionary of ItemObject from PyArrow Table

        Args:
            table (dict[str, Any]): PyArrow table
            schema (pa.schema): PyArrow schema
            source_id (str): Objects source ID

        Returns:
            dict[str, ItemObject]: Dictionary of ItemObject
        """

        items = table.to_pylist()
        objects = {}

        # Iterate on objects
        for index, item in enumerate(items):
            # Create object
            obj = ItemObject(
                id=item["id"],
                item_id=item["item_id"],
                view_id=item["view_id"],
                source_id=source_id,
            )

            # Add other base fields (review state, bbox, mask)
            for field in schema:
                if field.name == "review_state" and item["review_state"]:
                    obj.review_state = item["review_state"]
                if field.name == "bbox" and item["bbox"]:
                    obj.bbox = ItemBBox.from_pyarrow(item["bbox"])
                elif field.name == "mask" and item["mask"]:
                    obj.mask = ItemURLE.from_pyarrow(item["mask"])
            # Add features
            obj.features = ItemFeature.from_pyarrow(table.take([index]), schema)
            # Append object
            objects[item["id"]] = obj

        return objects

    def to_pyarrow(self) -> dict[str, Any]:
        """Return ItemObject in PyArrow format

        Returns:
            dict[str, Any]: Object in PyArrow format
        """

        pyarrow_object = {}

        # IDs
        pyarrow_object["id"] = self.id
        pyarrow_object["item_id"] = self.item_id
        pyarrow_object["view_id"] = self.view_id

        # Base fields (review state, bbox, mask)
        if self.review_state is not None:
            pyarrow_object["review_state"] = self.review_state

        pyarrow_mask = self.mask.to_pyarrow() if self.mask else None
        pyarrow_bbox = (
            self.bbox.to_pyarrow()
            if self.bbox
            else BBox.from_rle(pyarrow_mask) if pyarrow_mask else None
        )

        pyarrow_object["mask"] = pyarrow_mask.to_dict() if pyarrow_mask else None
        pyarrow_object["bbox"] = pyarrow_bbox.to_dict() if pyarrow_bbox else None

        # Features
        if self.features is not None:
            # Add features
            for feat in self.features.values():
                pyarrow_object[feat.name] = (
                    field_to_python(feat.dtype)(feat.value)
                    if feat.value is not None
                    else None
                )

            # Check feature types
            for feat in self.features.values():
                if pyarrow_object[feat.name] is not None and not isinstance(
                    pyarrow_object[feat.name], field_to_python(feat.dtype)
                ):
                    raise ValueError(
                        f"Feature {feat.name} of object {self.id} is of type {type(self.features[feat.name].value)} instead of type {field_to_python(feat.dtype)}"
                    )

        return pyarrow_object

    def add_or_update(
        self,
        ds_table: lancedb.db.LanceTable,
    ):
        """Add or update item object

        Args:
            ds_table (lancedb.db.LanceTable): Object table
        """

        # Convert object to PyArrow
        pyarrow_obj = self.to_pyarrow()
        table_obj = pa.Table.from_pylist(
            [pyarrow_obj],
            schema=ds_table.schema,
        )

        # Delete object (if it exists)
        scanner = ds_table.to_lance().scanner(filter=f"id in ('{self.id}')")
        existing_obj = scanner.to_table()
        if existing_obj.num_rows > 0:
            ds_table.delete(f"id in ('{self.id}')")

        # Add object
        ds_table.add(table_obj, mode="append")

        # Clear change history to prevent dataset from becoming too large
        ds_table.to_lance().cleanup_old_versions()



# @Copyright: CEA-LIST/DIASI/SIALV/LVA (2023)
# @Author: CEA-LIST/DIASI/SIALV/LVA <pixano@cea.fr>
# @License: CECILL-C
#
# This software is a collaborative computer program whose purpose is to
# generate and explore labeled data for computer vision applications.
# This software is governed by the CeCILL-C license under French law and
# abiding by the rules of distribution of free software. You can use,
# modify and/ or redistribute the software under the terms of the CeCILL-C
# license as circulated by CEA, CNRS and INRIA at the following URL
#
# http://www.cecill.info

import os
import re
from pathlib import Path


def natural_key(string: str) -> list:
    """Return key for string natural sort

    Args:
        string (str): Input string

    Returns:
        list: Sort key
    """

    return [int(s) if s.isdecimal() else s for s in re.split(r"(\d+)", string)]


def estimate_size(folder_path: Path) -> str:
    """Estimate folder size and return it as a human-readable string

    Args:
        folder_path (Path): Folder path

    Returns:
        str: Folder size as a human-readable string
    """

    # Estimate size
    total_size = 0
    for dirpath, _, filenames in os.walk(folder_path):
        for f in filenames:
            fp = os.path.join(dirpath, f)
            # skip if it is symbolic link
            if not os.path.islink(fp):
                total_size += os.path.getsize(fp)

    # Format size
    i = 0
    suffixes = ["B", "KB", "MB", "GB", "TB", "PB"]
    while total_size >= 1024 and i < len(suffixes) - 1:
        total_size /= 1024.0
        i += 1
    f = (f"{total_size:.2f}").rstrip("0").rstrip(".")
    readable_size = f"{f} {suffixes[i]}"

    return readable_size
